import logging
import re
from functools import cached_property, wraps
from pathlib import Path
from ssl import SSLContext
from typing import Optional, Any, Union, Literal, Mapping, OrderedDict, Iterator, Generator

from dotenv import find_dotenv
from httpx import Client, BaseTransport, URL, Response
from httpx._client import EventHook
from httpx._types import CertTypes
from pydantic import ConfigDict, Field, InstanceOf, FilePath, BaseModel
from pydantic_settings.sources import DotenvType

from ipfabric.auth import Setup, ProxyTypes, TimeoutTypes
from ipfabric.models import Snapshot, OAS, Endpoint, Snapshots, Methods, User, create_snapshot
from ipfabric.models.snapshot import ScheduledSnapshot
from ipfabric.settings import SeedList, Networks
from ipfabric.settings.authentication import CredentialList, PrivilegeList
from ipfabric.tools.configuration import trigger_backup
from ipfabric.tools.shared import VALID_REFS, raise_for_status, VALID_IP

logger = logging.getLogger("ipfabric")

LAST_ID, PREV_ID, LASTLOCKED_ID = VALID_REFS
RE_PATH = re.compile(r"^/?(api/)?v\d(\.\d+)?/")


def check_deprecated(http_method: str = None):
    def decorator_check_deprecated(func):
        @wraps(func)
        def wrapper_check_deprecated(self: "IPFabricAPI", *f_args, **f_kwargs):
            if method := http_method:
                url = f_kwargs.get("url", f_args[0])
            else:
                method = f_kwargs.get("method", f_args[0])
                url = f_kwargs.get("url", f_args[1])
            url = self._check_url(url)
            method = method if not url[:6].startswith("tables") else "post"
            _: Endpoint = getattr(getattr(self, "oas", {}).get(url, Methods(full_api_endpoint="")), method, None)
            if _ and _.deprecated:
                logger.warning(
                    f"API endpoint '{method.upper()} /{_.api_endpoint}' is marked deprecated in OpenAPI specification."
                )
            return func(self, *f_args, **f_kwargs)

        return wrapper_check_deprecated

    return decorator_check_deprecated


class IPFabricAPI(BaseModel):
    """Initializes the IP Fabric Client.

    Args:
        base_url: IP Fabric instance provided in 'base_url' parameter, or the 'IPF_URL' environment variable
        api_version: [Optional] Version of IP Fabric API
        snapshot_id: IP Fabric snapshot ID to use by default for database actions - defaults to '$last'
        auth: API token, tuple (username, password), or custom Auth to pass to httpx
        unloaded: True to load metadata from unloaded snapshots
        env_file: Path to .env file to load
        streaming: Default True to use streaming instead of paging.
        verify: httpx.Client - Default True to verify IPF SSL certificate.
        timeout: httpx.Client - Default 5
        proxy: httpx.Client
        mounts: httpx.Client
        cert: httpx.Client
        event_hooks: httpx.Client
        local_oas: Default True, False mainly for development purposes
        local_oas_file: Default True, False mainly for development purposes
        debug: Enable Debug.
    """

    model_config = ConfigDict(arbitrary_types_allowed=True)

    base_url: Optional[Union[str, InstanceOf[URL]]] = Field(
        None, description="Base URL of the IPF instance", examples=["https://demo.ipfabric.io"]
    )
    auth: Optional[Any] = Field(default=None, exclude=True)
    snapshot_id: Optional[str] = Field(None, description="Defaults to '$last'.")
    env_file: Optional[DotenvType] = Field(default=None, exclude=True)
    unloaded: bool = Field(False, description="Default False, do not load unloaded snapshot metadata.")
    streaming: bool = Field(True, description="Default True; use streaming results instead of pagination.")

    # HTTPX Options:
    verify: Optional[Union[SSLContext, str, bool]] = Field(default=None, exclude=True)
    timeout: Optional[Union[TimeoutTypes, Literal["DEFAULT"]]] = Field(default="DEFAULT", exclude=True)
    proxy: Optional[ProxyTypes] = Field(default=None, exclude=True)
    mounts: Optional[Mapping[str, Optional[BaseTransport]]] = Field(default=None, exclude=True)
    cert: Optional[CertTypes] = Field(default=None, exclude=True)
    event_hooks: Optional[Mapping[str, list[EventHook]]] = Field(default=None, exclude=True)
    http2: Optional[bool] = Field(default=True, exclude=True)

    # Debug/Other less used
    nvd_api_key: Optional[str] = None
    debug: bool = False
    api_version: Optional[str] = Field(None, description="Defaults to SDK or API version.")
    local_oas: bool = Field(True, description="Default True, use local minified OAS file instead of servers.")
    local_oas_file: Optional[FilePath] = None
    _os_version: Optional[str] = None
    _os_api_version: Optional[str] = None
    _client: Client = None
    _prev_snapshot_id: Optional[str] = None
    _attribute_filters: Optional[dict] = None
    _no_loaded_snapshots: bool = False
    _oas: Optional[OAS] = None
    _user: Optional[User] = None
    _snapshots: Optional[Snapshots] = None

    def model_post_init(self, __context: Any) -> None:
        env_file = (
            self.env_file if self.env_file else find_dotenv(usecwd=True) or Path("~").expanduser().joinpath(".env")
        )
        logger.info(f"Using .env file located at {env_file}.")
        setup = Setup(
            base_url=self.base_url,
            api_version=self.api_version,
            auth=self.auth,
            _env_file=env_file,
            snapshot_id=self.snapshot_id,
            verify=self.verify,
            timeout=self.timeout,
            proxy=self.proxy,
            mounts=self.mounts,
            cert=self.cert,
            event_hooks=self.event_hooks,
            debug=self.debug,
            http2=self.http2,
        )
        self._client = setup.client
        [setattr(self, k, v) for k, v in setup.update_attrs.items()]

        self._oas = OAS(client=self, local_oas=self.local_oas, local_oas_file=self.local_oas_file)
        self._user = self.get_user()
        self._snapshots = Snapshots(client=self)
        self.snapshot_id = setup.snapshot_id
        logger.debug(
            f"Successfully connected to '{self.base_url.host}' IPF version '{self.os_version}' "
            f"as user '{self.user.username}'"
        )

    def __setattr__(self, name: str, value: Any) -> None:
        if name == "timeout":
            self._client.timeout = value
        elif name == "snapshot_id":
            value = self._switch_snapshot(value)
        super().__setattr__(name, value)

    def _switch_snapshot(self, snapshot_id: str) -> Union[str, None]:
        """Verify snapshot ID is valid."""
        if not self.loaded_snapshots:
            logger.warning("No Snapshots are currently loaded.  Please load a snapshot before querying any data.")
            self._no_loaded_snapshots = True
            return None
        if snapshot_id in self.snapshots:
            self._prev_snapshot_id = self.snapshot_id
            return self.snapshots[snapshot_id].snapshot_id
        raise ValueError(f"Incorrect Snapshot ID: '{snapshot_id}'")

    @check_deprecated("get")
    def get(self, url, *args, params=None, **kwargs) -> Response:
        return self._client.get(url, *args, params=params, **kwargs)

    @check_deprecated("post")
    def post(self, url, *args, json=None, **kwargs) -> Response:
        return self._client.post(url, *args, json=json, **kwargs)

    @check_deprecated("put")
    def put(self, url, *args, json=None, **kwargs) -> Response:
        return self._client.put(url, *args, json=json, **kwargs)

    @check_deprecated("patch")
    def patch(self, url, *args, json=None, **kwargs) -> Response:
        return self._client.patch(url, *args, json=json, **kwargs)

    def request(self, method, url, *args, json=None, **kwargs) -> Response:
        return self._client.request(method, url, *args, json=json, **kwargs)

    @check_deprecated("delete")
    def delete(self, url, *args, **kwargs) -> Response:
        return self._client.delete(url, *args, **kwargs)

    def stream(self, method, url, *args, **kwargs) -> Iterator[Response]:
        return self._client.stream(method, url, *args, **kwargs)

    def _check_url(self, url) -> str:
        """Optimized URL processing with compiled regex and early returns"""

        path = URL(url).path
        if not path.startswith("/"):
            path = "/" + url

        # Early return for known web-to-api mappings
        if self._oas and path in self.web_to_api:
            return self.web_to_api[path].api_endpoint

        # Single regex match instead of multiple operations
        if match := RE_PATH.search(path):
            return path[match.end() :].lstrip("/")  # noqa: E203
        return path.lstrip("/")

    @cached_property
    def os_version(self):
        return self._os_version

    @cached_property
    def os_api_version(self):
        return self._os_api_version

    @cached_property
    def user(self) -> User:
        return self._user

    @cached_property
    def _api_insuf_rights(self):
        msg = f'API_INSUFFICIENT_RIGHTS for user "{self._user.username}" '
        if self._user.token:
            msg += f'token "{self._user.token.description}" '
        return msg

    @cached_property
    def hostname(self) -> str:
        resp = self.get("/os/hostname")
        if resp.status_code == 200:
            return resp.json()["hostname"]
        else:
            logger.critical(self._api_insuf_rights + 'on GET "/os/hostname"; Using URL host.')
            return str(self.base_url.host)

    @cached_property
    def oas(self) -> dict[str, Methods]:
        return self._oas.oas

    @cached_property
    def web_to_api(self) -> dict[str, Endpoint]:
        return self._oas.web_to_api

    @cached_property
    def scope_to_api(self) -> dict[str, Endpoint]:
        return self._oas.scope_to_api

    @property
    def attribute_filters(self):
        return self._attribute_filters

    @attribute_filters.setter
    def attribute_filters(self, attribute_filters: Union[dict[str, list[str]], None]):
        if attribute_filters:
            logger.warning(
                "Setting Global Attribute Filter for all tables/diagrams until explicitly unset to `None` or "
                f"overwritten in the method.\nFilter: {attribute_filters}"
            )
        self._attribute_filters = attribute_filters

    @property
    def snapshots(self) -> OrderedDict[str, Snapshot]:
        return self._snapshots.snapshots

    @property
    def snapshot(self) -> Snapshot:
        return self.snapshots[self.snapshot_id]

    @property
    def loaded_snapshots(self) -> OrderedDict[str, Snapshot]:
        """get only loaded snapshots"""
        return self._snapshots.loaded_snapshots

    @property
    def unloaded_snapshots(self) -> OrderedDict[str, Snapshot]:
        return self._snapshots.unloaded_snapshots

    @property
    def loading_snapshot(self) -> Union[Snapshot, None]:
        """Return Loading Snapshot or None"""
        return self._snapshots.loading_snapshot

    @property
    def running_snapshot(self) -> Union[Snapshot, None]:
        """Return Running Snapshot"""
        return self._snapshots.running_snapshot

    def get_snapshot(self, snapshot_id: str) -> Snapshot:
        return self._snapshots.get_snapshot(snapshot_id)

    def get_snapshot_id(self, snapshot: Union[Snapshot, str]) -> str:
        return self._snapshots.get_snapshot_id(snapshot)

    def get_snapshots(self) -> OrderedDict[str, Snapshot]:
        return self._snapshots.get_snapshots()

    @property
    def prev_snapshot_id(self) -> str:
        """get previous snapshot Id"""
        return self._prev_snapshot_id

    def get_user(self) -> User:
        """Gets current logged in user information.

        Returns:
            User: User model of logged in user
        """
        resp = raise_for_status(self.get("users/me"))
        user = User(**resp.json())
        if not (user.is_admin and user.token):  # TODO: NIM-14008: Implement logic after obtaining scopes for tokens
            resp = raise_for_status(self.get("users/me/scopes/api"))
            user.scopes = resp.json()["data"]
        return user

    def _ipf_pager(
        self,
        url: str,
        payload: dict,
        limit: int = 1000,
        start: int = 0,
    ) -> Generator:
        """
        Loops through and collects all the data from the tables
        :param url: str: Full URL to post to
        :param payload: dict: Data to submit to IP Fabric
        :param start: int: Where to start for the data
        :return: Generator: List of dictionaries
        """
        payload.setdefault("pagination", {})
        payload["pagination"]["limit"] = limit

        while True:
            payload["pagination"]["start"] = start
            response = raise_for_status(self.post(url, json=payload))
            chunk = response.json()["data"]

            if not chunk:
                break

            yield from chunk

            if len(chunk) < limit:
                break

            start += limit

    def trigger_backup(self, sn: str = None, ip: str = None):
        return trigger_backup(self, sn=sn, ip=ip)

    def create_snapshot(
        self,
        snapshot_name: str = "",
        snapshot_note: str = "",
        networks: Optional[Union[Networks, dict[str, list[Union[str, VALID_IP]]]]] = None,
        seeds: Optional[Union[SeedList, list[Union[str, VALID_IP]]]] = None,
        credentials: Optional[Union[CredentialList, list[dict]]] = None,
        privileges: Optional[Union[PrivilegeList, list[dict]]] = None,
        disabled_assurance_jobs: Optional[list[Literal["graphCache", "historicalData", "intentVerification"]]] = None,
        fail_if_running_snapshot: bool = True,
        **kwargs,
    ) -> ScheduledSnapshot:
        return create_snapshot(
            self,
            snapshot_name=snapshot_name,
            snapshot_note=snapshot_note,
            networks=networks,
            seeds=seeds,
            credentials=credentials,
            privileges=privileges,
            disabled_assurance_jobs=disabled_assurance_jobs,
            fail_if_running_snapshot=fail_if_running_snapshot,
            **kwargs,
        )



import json
import logging
from collections import OrderedDict
from ipaddress import (
    ip_interface,
    ip_address,
    IPv4Address,
    IPv6Address,
    IPv4Interface,
    IPv6Interface,
)
from typing import Any, Union, Literal

from macaddress import EUI48
from pydantic import BaseModel, Field, PrivateAttr

from ipfabric.models.oas import Endpoint, NestedColumn, Column
from ipfabric.tools.shared import parse_mac

LOGGER = logging.getLogger("ipfabric")

VM_INTS = "/tables/cloud/endpoints/virtual-machines-interfaces"
SEARCH_PATHS_IGNORE = [
    "/tables/networks",  # Deprecated
    "/tables/networks/ipv6-routes",  # Route tables match too many items
    "/tables/networks/routes",
    "/tables/mpls/l3-vpn/pe-routes",
    "/tables/routing/protocols/lisp/ipv4-routes",
    "/tables/mpls/forwarding",
    "/tables/addressing/path-lookup-sources",
    "/tables/addressing/path-lookup-sources-multicast",
    "/tables/addressing/path-lookup-sources-unicast",
    "/tables/management/changes/managed-devs",
    "/tables/reports/discovery-tasks",
    "/tables/interfaces/connectivity-matrix/unmanaged-neighbors/summary",  # Summary table
    "/tables/routing/protocols/bgp/address-families",
    "/tables/routing/protocols/ospf/interfaces",
    "/tables/routing/protocols/ospf-v3/interfaces",
    "/tables/management/connectivity-errors",
    "/tables/settings/jumphosts",
    "/tables/settings/ports",
    "/tables/networks/route-stability",
    "/tables/networks/gateway-redundancy",
]
RANKED = {
    "ipv4": [
        "/tables/addressing/managed-devs",
        "/tables/inventory/interfaces",
        "/tables/addressing/hosts",
        "/tables/addressing/arp",
        "/tables/addressing/ipv4-managed-ip-summary",
        "/tables/wireless/clients",
        "/tables/inventory/phones",
        VM_INTS,
        "/tables/neighbors/unmanaged",
        "/tables/interfaces/connectivity-matrix/unmanaged-neighbors/detail",
        "/tables/routing/protocols/bgp/neighbors",
        "/tables/routing/protocols/eigrp/neighbors",
        "/tables/routing/protocols/is-is/neighbors",
        "/tables/routing/protocols/ospf/neighbors",
        "/tables/routing/protocols/rip/neighbors",
    ],
    "ipv6": [
        "/tables/addressing/ipv6-managed-devs",
        "/tables/addressing/ipv6-hosts",
        "/tables/addressing/ipv6-neighbors",
        "/tables/neighbors/unmanaged",
        "/tables/interfaces/connectivity-matrix/unmanaged-neighbors/detail",
        VM_INTS,
        "/tables/routing/protocols/bgp/neighbors",
        "/tables/routing/protocols/is-is/neighbors",
        "/tables/routing/protocols/ospf-v3/neighbors",  # NIM-15216
    ],
    "mac": [
        "/tables/addressing/managed-devs",
        "/tables/addressing/ipv6-managed-devs",
        "/tables/inventory/interfaces",
        "/tables/addressing/hosts",
        "/tables/addressing/ipv6-hosts",
        "/tables/addressing/ipv6-neighbors",
        "/tables/addressing/arp",
        VM_INTS,
        "/tables/addressing/mac",
        "/tables/wireless/radio",
    ],
}
OPERATORS = {
    "CIDR": "cidr",
    "=": "eq",
    "IP": "ip",
    ">": "gt",
    ">=": "gte",
    "<": "lt",
    "<=": "lte",
    "@": "sect",
    "!@": "nsect",
}
VALID_OPS = Literal[
    "=", "IP", ">", ">=", "<", "<=", "@", "!@", "eq", "ip", "gt", "gte", "lt", "lte", "sect", "nsect", "cidr", "CIDR"
]
IP_ADDRESS = (IPv4Address, IPv6Address)
IP_INTERFACE = (IPv4Interface, IPv6Interface)


class GlobalSearch(BaseModel):
    client: Any = Field(exclude=True)
    _ipv4: list[Endpoint] = PrivateAttr(default_factory=list)
    _ipv6: list[Endpoint] = PrivateAttr(default_factory=list)
    _mac: list[Endpoint] = PrivateAttr(default_factory=list)

    def model_post_init(self, __context: Any) -> None:
        self._load_default()
        self._load_dynamic()

    def _load_default(self):
        for k, v in RANKED.items():
            for path in v:
                if getattr(self.client.oas[path[1:]].post, k).columns:
                    getattr(self.client.oas[path[1:]].post, k).full_scan = False
                    getattr(self, k).append(self.client.oas[path[1:]].post)

    def _load_dynamic(self):
        for path in self.client.oas.values():
            if path.post:
                if path.full_api_endpoint in SEARCH_PATHS_IGNORE:
                    continue
                for k, paths in RANKED.items():
                    if path.full_api_endpoint not in paths and getattr(path.post, k).columns:
                        getattr(self, k).append(path.post)

    @property
    def ipv4(self):
        return self._ipv4

    @property
    def ipv6(self):
        return self._ipv6

    @property
    def mac(self):
        return self._mac

    @staticmethod
    def _create_filter(
        columns: list[Union[Column, NestedColumn]], path: Endpoint, value: str, regex: bool = False
    ) -> dict:
        oper = "reg" if regex else "eq"
        filters = [
            (
                {col.parent: ["any", col.child, oper, value]}
                if isinstance(col, NestedColumn)
                else {col.name: ["any", oper, value]} if col.name in path.array_columns else {col.name: [oper, value]}
            )
            for col in columns
            if not col.filter.startswith("routing")
        ]
        return {"or": filters} if filters else None

    def _search(
        self,
        search_type: Literal["ipv4", "ipv6", "mac"],
        address: str,
        full_scan: bool = False,
        regex: bool = False,
        first_match: bool = False,
    ) -> dict[str, dict[str, Union[str, list]]]:

        results, data = OrderedDict(), list()

        for path in getattr(self, search_type):
            search_data = getattr(path, search_type)
            if first_match and data:
                LOGGER.debug("Found first match, stopping search.")
                break
            elif not full_scan and search_data.full_scan is True:
                LOGGER.debug('Finished searching default tables; "--full-scan" not enabled to search all tables.')
                break  # List is ordered and once this is met we can break if not full scan
            filters = self._create_filter(search_data.columns, path, address, True if regex else False)
            if not filters:
                continue
            msg = f'earching "{path.web_menu}": API Endpoint: "{path.full_api_endpoint}"'
            LOGGER.debug(f'S{msg}; Filters `{json.dumps(filters, separators=(",", ":"))}`.')
            data = self.client.fetch_all(path.api_endpoint, filters=filters)
            results[path.full_api_endpoint] = {
                "data": data,
                "path": path.full_api_endpoint,
                "webPath": path.web_endpoint,
                "url": path.filter_url(filters, self.client._client.base_url),
                "menu": f'{path.web_menu}: "{path.title or path.summary}"',
            }
            LOGGER.debug(f"Finished s{msg}.")

        return results

    def search(
        self,
        address: str,
        full_scan: bool = False,
        first_match: bool = False,
    ) -> Union[None, dict[str, dict[str, Union[str, list]]]]:

        if isinstance(address, int):
            raise TypeError(f"Input must be a valid string not integer: {str(address)}")
        try:
            return self.search_mac(address, full_scan, first_match)
        except ValueError:
            pass
        try:
            return self.search_ip(address, full_scan, first_match)
        except ValueError:
            raise SyntaxError(f'Address does not appear to be a IPv4, IPv6, nor MAC Address: "{address}".')

    def search_mac(
        self,
        address: str,
        full_scan: bool = False,
        first_match: bool = False,
    ) -> dict[str, dict[str, Union[str, list]]]:

        if isinstance(address, int):
            raise TypeError(f"Input must be a valid string not integer: {str(address)}")
        LOGGER.info("Verifying Address is a MAC.")
        EUI48(address)
        mac = parse_mac(address)
        LOGGER.debug(f'Searching for MAC Address "{mac}".')
        return self._search("mac", mac, full_scan, first_match=first_match)

    def search_ip(
        self,
        address: str,
        full_scan: bool = False,
        first_match: bool = False,
    ) -> dict[str, dict[str, Union[str, list]]]:
        LOGGER.info("Verifying Address is an IP.")
        ip = ip_address(address)
        if ip.version == 4:
            LOGGER.debug(f'Searching for IPv4 Address "{str(ip)}".')
            return self._search("ipv4", str(ip), full_scan, first_match=first_match)
        elif ip.version == 6:
            LOGGER.debug(f'Searching for IPv6 Address "{str(ip)}".')
            return self._search("ipv6", str(ip), full_scan, first_match=first_match)

    def _search_ip(
        self,
        version: int,
        address: str,
        full_scan: bool = False,
        first_match: bool = False,
    ) -> dict[str, dict[str, Union[str, list]]]:
        ip = ip_address(address)
        if ip.version != version:
            raise ValueError()
        return self._search("ipv" + str(version), str(ip), full_scan, first_match=first_match)

    def search_ipv4(
        self,
        address: str,
        full_scan: bool = False,
        first_match: bool = False,
    ) -> Union[None, dict[str, dict[str, Union[str, list]]]]:
        return self._search_ip(4, address, full_scan, first_match)

    def search_ipv6(
        self,
        address: str,
        full_scan: bool = False,
        first_match: bool = False,
    ) -> Union[None, dict[str, dict[str, Union[str, list]]]]:
        return self._search_ip(6, address, full_scan, first_match)

    def search_regex(
        self,
        search_type: Literal["ipv4", "ipv6", "mac"],
        address: str,
        full_scan: bool = False,
        first_match: bool = False,
    ) -> dict[str, dict[str, Union[str, list]]]:
        LOGGER.debug(f'Searching for {search_type.upper()} Address "{address}".')
        return self._search(search_type, address, full_scan, True, first_match)


class RouteTableSearch(BaseModel):
    client: Any = Field(exclude=True)
    _route: str = "tables/networks/routes"
    _route6: str = "tables/networks/ipv6-routes"

    @staticmethod
    def _check_ip(address: str) -> Union[IPv4Address, IPv6Address, IPv4Interface, IPv6Interface]:
        try:
            return ip_address(address)
        except ValueError:
            LOGGER.debug(f'Address "{address}" is not IP Address, trying as subnet.')
            try:
                return ip_interface(address)
            except ValueError:
                raise SyntaxError(f'Address does not appear to be a IPv4 or IPv6 Address/Subnet: "{address}".')

    @staticmethod
    def _check_search(
        ip: Union[IPv4Address, IPv6Address, IPv4Interface, IPv6Interface],
        operator: str,
        next_hop: bool = False,
    ) -> Union[IPv4Address, IPv6Address, IPv4Interface, IPv6Interface]:
        if next_hop:
            if operator not in ["cidr", "eq"]:
                raise SyntaxError("IP, EQ, or CIDR search only allowed for Next Hop IP.")
            elif isinstance(ip, IP_INTERFACE) and operator == "eq":
                ip = ip.ip
        else:
            if ip.version == 6 and operator not in ["ip", "eq"]:
                raise SyntaxError('IPv6 CIDR Search not implemented yet, please use "ip" or "eq".')  # NIM-14175
            elif operator == "cidr":
                raise SyntaxError('"CIDR" Search only used in Next Hop IP.')
            elif isinstance(ip, IP_INTERFACE) and operator == "ip" and ip.max_prefixlen != ip.network.prefixlen:
                raise SyntaxError(f'IP Search must be IP not CIDR: "{str(ip)}".')
            elif isinstance(ip, IP_ADDRESS) and operator != "ip":
                ip = ip_interface(str(ip))
            elif isinstance(ip, IP_INTERFACE) and operator == "ip":
                ip = ip.ip
        return ip

    def search(
        self, address: str, operator: VALID_OPS = "=", next_hop: bool = False, ignore_default: bool = True
    ) -> dict[str, Any]:
        """ """
        if operator not in OPERATORS.values():
            operator = OPERATORS[operator]
        operator = "eq" if operator == "ip" and next_hop else operator
        ip = self._check_search(self._check_ip(address), operator, next_hop)

        filters = {"nexthop": ["any", "ip", operator, str(ip)]} if next_hop else {"network": [operator, str(ip)]}
        if ignore_default and str(ip) not in ["0.0.0.0/0", "::/0"]:
            filters = {"and": [filters, {"network": ["neq", "0.0.0.0/0" if ip.version == 4 else "::/0"]}]}
        api_endpoint = self._route if ip.version == 4 else self._route6
        path = self.client.oas[api_endpoint].post
        return {
            "data": self.client.fetch_all(url=api_endpoint, filters=filters),
            "path": path.full_api_endpoint,
            "webPath": path.web_endpoint,
            "url": path.filter_url(filters, self.client._client.base_url),
            "menu": f'{path.web_menu}: "{path.title or path.summary}"',
        }



from .config_export import ExportConfig
from .device import Device, Devices
from .extensions import Extensions
from .global_search import GlobalSearch, RouteTableSearch
from .intent import Intent
from .inventory import Inventory
from .jobs import Jobs, Job
from .oas import OAS, Endpoint, Methods
from .rbac import Role, Policy
from .security import SecurityModel
from .snapshot import Snapshot, SNAPSHOT_COLUMNS, create_snapshot, snapshot_upload
from .snapshots import Snapshots
from .table import BaseTable, Table
from .technology import Technology
from .users import User

__all__ = [
    "BaseTable",
    "Table",
    "Inventory",
    "Technology",
    "Jobs",
    "Job",
    "Snapshot",
    "Intent",
    "SNAPSHOT_COLUMNS",
    "Device",
    "Devices",
    "OAS",
    "Endpoint",
    "Methods",
    "Snapshots",
    "User",
    "Role",
    "Policy",
    "GlobalSearch",
    "RouteTableSearch",
    "create_snapshot",
    "snapshot_upload",
    "SecurityModel",
    "Extensions",
    "ExportConfig",
]



import base64
import logging
from time import sleep
from typing import Any, Literal, Optional, Union

import httpx
from pydantic import BaseModel, Field, field_validator, ConfigDict

from ipfabric.tools.shared import raise_for_status, VALID_REFS, valid_snapshot
from .table import BaseTable

logger = logging.getLogger("ipfabric")

SNAP_JOBS = {
    "load": "snapshotLoad",
    "unload": "snapshotUnload",
    "download": "snapshotDownload",
    "add": "discoveryAdd",
    "refresh": "discoveryRefresh",
    "delete": "deleteDevice",
    "recalculate": "recalculateSites",
    "new": "discoveryNew",
}
ALL_JOBS = {
    "export": "configurationExport",
    "import": "configurationImport",
    "load_graph": "loadGraphCache",
    "history": "saveHistoricalData",
    "report": "report",
    **SNAP_JOBS,
}

SNAP_ACTIONS = Literal["load", "unload", "download", "add", "refresh", "delete", "discoveryNew"]
ALL_ACTIONS = Union[Literal["export", "import"], SNAP_ACTIONS]
SORT = {"order": "desc", "column": "startedAt"}


class TechsupportSnapshotSettings(BaseModel):
    id: Union[str, Literal["$last", "$prev", "$lastLocked"]] = "$last"
    backupDb: bool = True
    removeCli: bool = False

    @field_validator("id")
    @classmethod
    def _valid_snapshot(cls, v: str) -> str:
        return valid_snapshot(v)


class TechsupportPayload(BaseModel):
    databases: bool = False
    systemLogs: bool = True
    discoveryServicesLogs: bool = True
    snapshot: TechsupportSnapshotSettings = Field(default_factory=TechsupportSnapshotSettings)
    usageData: bool = True
    ipfChecker: bool = True


class Job(BaseModel):
    model_config = ConfigDict(extra="allow")
    finishedAt: Optional[int] = None
    snapshot: Optional[str] = None
    name: Optional[str] = None
    id: Optional[str] = None
    username: Optional[str] = None
    isDone: bool = False
    scheduledAt: Optional[int] = None
    downloadFile: Optional[str] = None
    startedAt: Optional[int] = None
    status: Optional[str] = None

    @field_validator("snapshot")
    @classmethod
    def _empty_str_to_none(cls, v: Union[None, str]) -> Union[None, str]:
        return v if v else None


class Jobs(BaseModel):
    client: Any = Field(exclude=True)

    @property
    def all_jobs(self):
        return BaseTable(client=self.client, endpoint="tables/jobs")

    @property
    def columns(self):
        return [
            "id",
            "downloadFile",
            "finishedAt",
            "isDone",
            "name",
            "scheduledAt",
            "snapshot",
            "startedAt",
            "status",
            "username",
        ]

    def get_job_by_id(self, job_id: Union[str, int]) -> Optional[Job]:
        """Get a job by its ID and returns it as a Job object.

        Args:
            job_id: ID of the job to retrieve

        Returns: Job object if found, None if not found

        """
        jobs = self.all_jobs.all(filters={"id": ["eq", str(job_id)]}, columns=self.columns)
        if not jobs:
            return None
        return Job(**jobs[0])

    def find_job(self, started: int, action: ALL_ACTIONS, snapshot_id: str = None) -> Union[Job, None]:
        """Finds a job and returns it.

        Args:
            started: int: Integer time since epoch in milliseconds
            action: str: Type of job to filter on.
            snapshot_id: str: Optional: UUID of a snapshot to filter on

        Returns:
            Job: Job object or None if not found.
        """
        j_filter = dict(name=["eq", ALL_JOBS[action]], startedAt=["gte", started - 1000])
        j_filter.update(dict(snapshot=["eq", snapshot_id]) if snapshot_id else dict())
        sleep(5)  # give the IPF server a chance to start the job
        # find the running snapshotDownload job (i.e. not done)
        jobs = self.all_jobs.fetch(
            filters=j_filter,
            sort={"order": "desc", "column": "startedAt"},
            columns=self.columns,
        )
        logger.debug(f"Job filter: {j_filter}\nList of jobs:{jobs}")
        if not jobs:
            logger.warning(f"Job not found: {j_filter}")
            return None
        return Job(**jobs[0])

    def return_job_when_done(self, job: Union[Job, str], retry: int = 5, timeout: int = 5) -> Union[Job, None]:
        """
        Returns the finished job.

        Args:
            job: str: Job ID string; Job: Job object
            retry: int: How many times to query the table
            timeout: int: How long to wait in-between retries

        Returns:
            job: Job: Object about the job
        """
        job = self.get_job_by_id(job if isinstance(job, str) else job.id)
        if not job:
            logger.critical(f"Job not found: {job}")
            return None

        if job.isDone:
            return job

        for _ in range(retry):
            job = self.get_job_by_id(job.id)
            if job.isDone:
                return job
            sleep(timeout)

        return None

    def check_snapshot_job(
        self, snapshot_id: str, started: int, action: SNAP_ACTIONS, retry: int = 5, timeout: int = 5
    ) -> Union[Job, None]:
        """Checks to see if a snapshot load job is completed.

        Args:
            snapshot_id: str: UUID of a snapshot
            started: int: Integer time since epoch in milliseconds
            action: str: Type of job to filter on
            timeout: int: How long in seconds to wait before retry
            retry: int: How many retries to use when looking for a job, increase for large downloads

        Returns:
            Job: Job object or None if did not complete.
        """
        job = self.find_job(started=started, action=action, snapshot_id=snapshot_id)
        return self.return_job_when_done(job, retry=retry, timeout=timeout)

    def check_snapshot_assurance_jobs(
        self, snapshot_id: str, assurance_settings: dict, started: int, retry: int = 5, timeout: int = 5
    ):
        """Checks to see if a snapshot Assurance Engine calculation jobs are completed.

        Args:
            snapshot_id: UUID of a snapshot
            assurance_settings: Dictionary from Snapshot.get_assurance_engine_settings
            started: Integer time since epoch in milliseconds
            timeout: How long in seconds to wait before retry
            retry: how many retries to use when looking for a job, increase for large downloads

        Returns:
            True if load is completed, False if still loading
        """
        if assurance_settings["disabled_graph_cache"] is False:
            job = self.find_job(started=started, action="load_graph", snapshot_id=snapshot_id)
            if not job or self.return_job_when_done(job, retry=retry, timeout=timeout) is None:
                logger.error("Graph Cache did not finish loading; Snapshot is not fully loaded yet.")
                return False
        if assurance_settings["disabled_historical_data"] is False:
            job = self.find_job(started=started, action="history", snapshot_id=snapshot_id)
            if not job or self.return_job_when_done(job, retry=retry, timeout=timeout) is None:
                logger.error("Historical Data did not finish loading; Snapshot is not fully loaded yet.")
                return False
        if assurance_settings["disabled_intent_verification"] is False:
            job = self.find_job(started=started, action="report", snapshot_id=snapshot_id)
            if not job or self.return_job_when_done(job, retry=retry, timeout=timeout) is None:
                logger.error("Intent Calculations did not finish loading; Snapshot is not fully loaded yet.")
                return False
        return True

    def generate_techsupport(
        self,
        payload: TechsupportPayload = TechsupportPayload(),
        wait_for_ts: bool = True,
        timeout: int = 60,
        retry: int = 5,
    ) -> Job:
        if payload.snapshot.id in VALID_REFS:
            payload.snapshot.id = self.client.snapshots[payload.snapshot.id].snapshot_id
        _ = raise_for_status(self.client.post(url="/os/techsupport", json=payload.model_dump()))

        # Wait for the job to start
        sleep(2)
        tech_support_jobs = self.all_jobs.all(filters={"name": ["eq", "techsupport"]}, sort=SORT)

        if not tech_support_jobs:
            raise LookupError("No techsupport jobs found.")

        job = tech_support_jobs[0]
        if wait_for_ts:
            job = self.return_job_when_done(
                job=job,
                retry=retry,
                timeout=timeout,
            )
        return job

    def download_job_file(self, job_id: str) -> httpx.Response:
        return self.client.get(f"jobs/{str(job_id)}/download")

    def download_techsupport_file(self, job_id: str) -> httpx.Response:
        return self.download_job_file(job_id)

    def upload_techsupport_file(
        self,
        upload_username: str = "techsupport",
        upload_password: Optional[str] = None,
        upload_file_timeout: int = 600,
        upload_server: Literal["eu", "us"] = "eu",
        techsupport_bytes: bytes = None,
        techsupport_job_id: str = None,
        upload_verify: bool = True,
    ):
        if not upload_password:
            raise ValueError("Upload password is required.")
        if not techsupport_job_id and not techsupport_bytes:
            raise ValueError("Techsupport bytes or Job ID is required.")

        if not techsupport_bytes:
            resp = self.download_techsupport_file(techsupport_job_id)
            if resp.status_code != 200:
                raise httpx.HTTPStatusError(
                    f"Failed to download techsupport file: {resp.status_code}", request=resp.request, response=resp
                )
            techsupport_bytes = resp.content
        base64_credentials = base64.b64encode(f"{upload_username}:{upload_password}".encode("utf-8")).decode("utf-8")

        headers = {
            "Content-Type": "application/x-tar",
            "Accept": "application/json",
            "Authorization": f"Basic {base64_credentials}",
            "Content-Length": str(len(techsupport_bytes)),
        }

        upload_response = httpx.post(
            url=f"https://upload.{upload_server}.ipfabric.io/upload",
            headers=headers,
            content=techsupport_bytes,
            timeout=upload_file_timeout,
            verify=upload_verify,
        )

        if upload_response.status_code == 200:
            logger.info("Successfully uploaded techsupport file")
        else:
            logger.error(f"Failed to upload techsupport file. Status code: {upload_response.status_code}")
            logger.error(f"Response content: {upload_response.text}")
            upload_response.raise_for_status()



import os
from typing import Optional, List, Union
from uuid import UUID

from pydantic import BaseModel, Field

from .protocols import PROTOCOLS
from .trace import Trace
from ..input_models import NetworkSettings, PathLookupSettings

PYDANTIC_EXTRAS = os.getenv("IPFABRIC_PYDANTIC_EXTRAS", "allow")


class Extra(BaseModel, extra=PYDANTIC_EXTRAS):
    ip: Optional[str] = None


class Line(BaseModel, extra=PYDANTIC_EXTRAS):
    pattern: Optional[str] = None
    color: Optional[int] = None
    thickness: Optional[int] = None


class Style(BaseModel, extra=PYDANTIC_EXTRAS):
    background: Optional[str] = None
    line: Optional[Line] = None


class Position(BaseModel, extra=PYDANTIC_EXTRAS):
    x: Union[int, float]
    y: Union[int, float]


class EdgePosition(BaseModel, extra=PYDANTIC_EXTRAS):
    c: Position
    e: Position


class Label(BaseModel, extra=PYDANTIC_EXTRAS):
    type: Optional[str] = None
    visible: Optional[bool] = None
    text: Optional[str] = None
    angle: Optional[int] = None
    anchor: Optional[Position] = None
    position: Optional[Position] = None
    wrapText: Optional[bool] = None


class Labels(BaseModel, extra=PYDANTIC_EXTRAS):
    center: Optional[Union[list[Label], Label]] = None
    source: Optional[Union[list[Label], Label]] = None
    target: Optional[Union[list[Label], Label]] = None


class ArrowHeads(BaseModel, extra=PYDANTIC_EXTRAS):
    target: Optional[list[Position]] = Field(default_factory=list)
    source: Optional[list[Position]] = Field(default_factory=list)


class Positions(BaseModel, extra=PYDANTIC_EXTRAS):
    line: Optional[list[Union[Position, EdgePosition]]] = Field(default_factory=list)
    arrowHeads: Optional[ArrowHeads] = None
    labels: Optional[Labels] = None


class Checks(BaseModel, extra=PYDANTIC_EXTRAS):
    green: int = Field(alias="0")
    blue: int = Field(alias="10")
    amber: int = Field(alias="20")
    red: int = Field(alias="30")


class Severity(Checks, extra=PYDANTIC_EXTRAS):
    pass


class Topics(BaseModel, extra=PYDANTIC_EXTRAS):
    acl: Checks = Field(alias="ACL")
    forwarding: Checks = Field(alias="FORWARDING")
    zonefw: Checks = Field(alias="ZONEFW")
    nat44: Optional[Checks] = Field(None, alias="NAT44")
    pbr: Optional[Checks] = Field(None, alias="PBR")


class TrafficScore(BaseModel, extra=PYDANTIC_EXTRAS):
    accepted: int
    dropped: int
    forwarded: int
    total: int


class Packets(BaseModel, extra=PYDANTIC_EXTRAS):
    packet: Optional[list[PROTOCOLS]] = Field(default_factory=list)
    ifaceName: Optional[str] = None
    prevEdgeIds: Optional[list[str]] = Field(default_factory=list)
    nextEdgeIds: Optional[list[str]] = Field(default_factory=list)
    severityInfo: Optional[Checks] = None
    trafficScore: Optional[TrafficScore] = None


class Node(BaseModel, extra=PYDANTIC_EXTRAS):
    path: Optional[str] = None
    boxId: Optional[str] = None
    children: List
    graphType: str
    id: str
    label: str
    parentPath: Optional[str] = None
    sn: str
    type: str
    stack: Optional[bool] = None
    position: Optional[Position] = None
    style: Optional[Style] = None
    acceptedPackets: Optional[dict[str, Packets]] = Field(default_factory=dict)
    droppedPackets: Optional[dict[str, Packets]] = Field(default_factory=dict)
    generatedPackets: Optional[dict[str, Packets]] = Field(default_factory=dict)
    extra: Optional[Extra] = None


class Edge(BaseModel, extra=PYDANTIC_EXTRAS):
    direction: str
    source: str
    target: str
    edgeSettingsId: UUID
    id: str
    labels: Labels
    protocol: Optional[str] = ""
    shift: Optional[Union[int, float]] = None
    positions: Optional[Positions] = None
    style: Optional[Style] = None


class RelatedTechnology(BaseModel, extra=PYDANTIC_EXTRAS):
    vlanId: int
    rootId: str


class NetworkEdge(Edge, BaseModel, extra=PYDANTIC_EXTRAS):
    circle: bool
    children: list[str]
    relatedTechnology: Optional[RelatedTechnology] = None


class PathLookupEdge(Edge, BaseModel, extra=PYDANTIC_EXTRAS):
    nextEdgeIds: list[str]
    prevEdgeIds: list[str]
    packet: list[PROTOCOLS]
    severityInfo: Severity
    sourceIfaceName: Optional[str] = None
    targetIfaceName: Optional[str] = None
    trafficScore: TrafficScore
    nextEdge: Optional[list] = Field(default_factory=list)
    prevEdge: Optional[list] = Field(default_factory=list)
    loopIds: Optional[list[int]] = Field(default_factory=list)


class EventsSummary(BaseModel, extra=PYDANTIC_EXTRAS):
    flags: list
    topics: Topics
    global_list: list = Field(alias="global")


class Traces(BaseModel, extra=PYDANTIC_EXTRAS):
    severityInfo: Checks
    sourcePacketId: str
    targetPacketId: str
    trace: list[Trace]


class Decision(BaseModel, extra=PYDANTIC_EXTRAS):
    traces: list[Traces]
    trafficIn: Optional[dict[str, list[str]]] = Field(default_factory=dict)
    trafficOut: Optional[dict[str, list[str]]] = Field(default_factory=dict)


class Check(BaseModel, extra=PYDANTIC_EXTRAS):
    exists: bool


class PathLookup(BaseModel, extra=PYDANTIC_EXTRAS):
    eventsSummary: EventsSummary
    decisions: dict[str, Decision]
    passingTraffic: str
    check: Check


class GraphData(BaseModel, extra=PYDANTIC_EXTRAS):
    nodes: dict[str, Node]
    edges: dict[str, Union[NetworkEdge, PathLookupEdge]]


class Entities(BaseModel, extra=PYDANTIC_EXTRAS):
    deviceTypes: list[str]
    edgeSettingsIds: list[str]


class Path(BaseModel, extra=PYDANTIC_EXTRAS):
    defaultLayout: str
    layoutAlgorithm: str


class Topology(BaseModel, extra=PYDANTIC_EXTRAS):
    availableEntities: Entities
    allPathsSelected: bool
    paths: dict[str, Path]


class GraphResult(BaseModel, extra=PYDANTIC_EXTRAS):
    boxLabels: dict[str, str]
    graphData: GraphData
    settings: Union[NetworkSettings, PathLookupSettings]


class BaseResult(BaseModel, extra=PYDANTIC_EXTRAS):
    graphResult: GraphResult

    @property
    def nodes(self) -> dict[str, Node]:
        return self.graphResult.graphData.nodes


class TopologyResult(BaseResult, BaseModel, extra=PYDANTIC_EXTRAS):
    topology: Topology

    @property
    def edges(self) -> dict[str, NetworkEdge]:
        return self.graphResult.graphData.edges


class PathLookupResult(BaseResult, BaseModel, extra=PYDANTIC_EXTRAS):
    pathlookup: Optional[PathLookup] = None

    @property
    def edges(self) -> dict[str, PathLookupEdge]:
        return self.graphResult.graphData.edges



import json
import logging
from datetime import datetime, timezone, timedelta
from typing import Any, Union

from pydantic import BaseModel, Field

from ipfabric.models.config_export import ExportConfig, ExportJob, ImportJob
from ipfabric.tools.shared import raise_for_status

logger = logging.getLogger("ipfabric")

CONFIG_OPTIONS = {
    "device-attributes",
    "jumphost",
    "oui",
    "discovery",
    "discovery-seeds",
    "device-credentials",
    "assurance-engine",
    "discovery-tasks",
    "vendors-api",
    "site-separation",
    "cli-settings",
    "bgp-routing",
    "snapshot-retention",
    "configuration-management",
    "api-tokens",
    "webhooks",
    "snmp",
    "certificates",
    "local-users",
    "ldap",
    "roles",
    "policies",
    "intent-checks",
    "dashboards-and-widgets",
}
CONFIG_ITEMS = set(sorted(CONFIG_OPTIONS.union({_.replace("-", "_") for _ in CONFIG_OPTIONS})))


class ApplianceConfiguration(BaseModel):
    client: Any = Field(description="IPFClient", exclude=True)

    @staticmethod
    def _create_options(
        passphrase: str,
        config_options: Union[ExportConfig, list[str], None],
        disable_options: Union[list[str], None] = None,
    ) -> ExportConfig:
        if len(passphrase) < 8:
            raise ValueError("Passphrase must be at least 8 characters long")
        if isinstance(config_options, list):
            if not all(item not in CONFIG_ITEMS for item in config_options):
                raise ValueError(f"Invalid configuration options: {set(config_options) - CONFIG_ITEMS}")
            config_options = ExportConfig(**{k: k in config_options for k in CONFIG_ITEMS})
        if config_options is None:
            config_options = ExportConfig()
        if disable_options is not None:
            if not all(item not in CONFIG_ITEMS for item in disable_options):
                raise ValueError(f"Invalid disabled configuration options: {set(disable_options) - CONFIG_ITEMS}")
            for k, v in ExportConfig(**{k: k not in disable_options for k in CONFIG_ITEMS}).model_dump().items():
                if v is False:
                    setattr(config_options, k, False)
        return config_options

    def export_config(
        self,
        passphrase: str,
        config_options: Union[ExportConfig, list[str], None] = None,
        disable_options: Union[list[str], None] = None,
        wait_for_export: bool = True,
        retry: int = 5,
        timeout: int = 5,
    ) -> Union[ExportJob, None]:
        """
        Export the current appliance configuration.

        Args:
            passphrase: str: 8 characters or longer string to encrypt the configuration.
            config_options: list: List of configuration options to export; None: export all; or ExportConfig object.
            disable_options: list: List of configuration options to disable export; None: export all
            wait_for_export: bool: Wait for the job to complete before returning
            timeout: int: How long in seconds to wait before retry
            retry: int: How many retries to use when looking for a job, increase for large downloads

        Returns: ExportJob: Job object with configuration data; None if an error occurred.
        """
        config_options = self._create_options(passphrase, config_options, disable_options)
        payload = {"passphrase": passphrase, **config_options.export()}
        started = round((datetime.now(timezone.utc) - timedelta(seconds=10)).timestamp() * 1000)
        raise_for_status(self.client.post("appliance-configuration/export", json=payload))
        # TODO: NIM-18675 v8.0 - added in 7.3
        job = self.client.jobs.find_job(started=started, action="export")
        if not job:
            return None
        if wait_for_export and (_ := self.client.jobs.return_job_when_done(job, retry=retry, timeout=timeout)):
            config = raise_for_status(self.client.get(f"jobs/{job.id}/download"))
            return ExportJob(job=_, config=config.json())
        return ExportJob(job=job, config=None)

    def import_config(
        self,
        passphrase: str,
        config: Union[dict, str, bytes],
        config_options: Union[ExportConfig, list[str], None] = None,
        disable_options: Union[list[str], None] = None,
        wait_for_import: bool = True,
        retry: int = 5,
        timeout: int = 5,
    ) -> Union[ImportJob, None]:
        """
        Export the current appliance configuration.

        Args:
            passphrase: str: 8 characters or longer string to encrypt the configuration.
            config: dict: Configuration data to import.
            config_options: list: List of configuration options to export; None: export all; or ExportConfig object.
            disable_options: list: List of configuration options to disable export; None: export all
            wait_for_import: bool: Wait for the job to complete before returning
            timeout: int: How long in seconds to wait before retry
            retry: int: How many retries to use when looking for a job, increase for large downloads

        Returns: ImportJob: Job object with configuration data; None if an error occurred.
        """
        config_options = self._create_options(passphrase, config_options, disable_options).export()[
            "configurationOptions"
        ]
        valid_options = set(config_options).intersection(set(config.keys()))
        if not valid_options:
            raise ValueError("No valid configuration options to be imported.")
        if valid_options != set(config_options):
            logger.warning(
                "The following settings were not in the configuration and will not be imported:\n"
                f"{json.dumps(list(set(config_options) - valid_options))}"
            )

        if isinstance(config, dict):
            config = json.dumps(config).encode("utf-8")
        config = config if isinstance(config, bytes) else config.encode("utf-8")
        files = [
            ("file", ("configurationExport.json", config, "application/json")),
            ("passphrase", (None, passphrase)),
            *[("configurationOptions[]", (None, v)) for v in valid_options],
        ]
        started = round((datetime.now(timezone.utc) - timedelta(seconds=10)).timestamp() * 1000)
        raise_for_status(self.client.post("appliance-configuration/import", files=files))
        # TODO: NIM-18675 v8.0
        job = self.client.jobs.find_job(started=started, action="import")
        if not job:
            return None
        if wait_for_import:
            _ = self.client.jobs.return_job_when_done(job, retry=retry, timeout=timeout)
            job = _ or self.client.jobs.get_job_by_id(job.id)
        return ImportJob(job=job)



"""
Python3 script to search for Devices in the Inventory Table.
"""

import argparse
import csv
import json
import logging
import os
import re
import sys
from collections import defaultdict
from typing import Union

from ipfabric import IPFClient
from ipfabric.scripts.shared import shared_args

try:
    from rich.console import Console
    from rich.table import Table
    from rich import box
except ImportError:
    raise ImportError("Rich is required for printing, please install by using `pip3 install rich`.")

CONSOLE = Console()

logging.basicConfig(format="%(levelname)s: %(message)s")

LOGGER = logging.getLogger("ipf_device_search")

VALID_COLUMNS = {
    "uptime": {"name": "uptime", "filter": "float", "alt_name": None},
    "model": {"name": "model", "filter": "str", "alt_name": None},
    "reload": {"name": "reload", "filter": "str", "alt_name": None},
    "image": {"name": "image", "filter": "str", "alt_name": None},
    "domain": {"name": "domain", "filter": "str", "alt_name": None},
    "platform": {"name": "platform", "filter": "str", "alt_name": None},
    "slug": {"name": "slug", "filter": "str", "alt_name": None},
    "hostname": {"name": "hostname", "filter": "str", "alt_name": None},
    "fqdn": {"name": "fqdn", "filter": "str", "alt_name": None},
    "processor": {"name": "processor", "filter": "str", "alt_name": None},
    "sn": {"name": "sn", "filter": "str", "alt_name": None},
    "version": {"name": "version", "filter": "str", "alt_name": None},
    "vendor": {"name": "vendor", "filter": "str", "alt_name": None},
    "family": {"name": "family", "filter": "str", "alt_name": None},
    "stpdomain": {"name": "stpDomain", "filter": "str", "alt_name": "stp_domain"},
    "hostnameoriginal": {"name": "hostnameOriginal", "filter": "str", "alt_name": "hostname_original"},
    "loginip": {"name": "loginIp", "filter": "str", "alt_name": "login_ip"},
    "snhw": {"name": "snHw", "filter": "str", "alt_name": "sn_hw"},
    "memorytotalbytes": {"name": "memoryTotalBytes", "filter": "float", "alt_name": "mem_total_bytes"},
    "hostnameprocessed": {"name": "hostnameProcessed", "filter": "str", "alt_name": "hostname_processed"},
    "sitename": {"name": "siteName", "filter": "str", "alt_name": "site"},
    "devtype": {"name": "devType", "filter": "str", "alt_name": "dev_type"},
    "tsdiscoveryend": {"name": "tsDiscoveryEnd", "filter": "float", "alt_name": "ts_discovery_end"},
    "configreg": {"name": "configReg", "filter": "str", "alt_name": "config_reg"},
    "rd": {"name": "rd", "filter": "str", "alt_name": "routing_domain"},
    "memoryusedbytes": {"name": "memoryUsedBytes", "filter": "float", "alt_name": "mem_used_bytes"},
    "memoryutilization": {"name": "memoryUtilization", "filter": "float", "alt_name": "mem_utilization"},
    "secdiscoveryduration": {"name": "secDiscoveryDuration", "filter": "float", "alt_name": "sec_discovery_duration"},
    "logintype": {"name": "loginType", "filter": "str", "alt_name": "login_type"},
    "tsdiscoverystart": {"name": "tsDiscoveryStart", "filter": "float", "alt_name": "ts_discovery_start"},
    "loginport": {"name": "loginPort", "filter": "int", "alt_name": "login_port"},
    "objectid": {"name": "objectId", "filter": "str", "alt_name": "object_id"},
    "taskkey": {"name": "taskKey", "filter": "str", "alt_name": "task_key"},
}
VALID_ALT_COLUMNS = {_["alt_name"]: _ for _ in VALID_COLUMNS.values()}
COLUMNS_HELP = [
    f"{_['name']}{'|' + _['alt_name'] if _['alt_name'] else ''} ({_['filter']})" for _ in VALID_COLUMNS.values()
]
DEFAULT_COLUMNS = [
    "hostname",
    "siteName",
    "vendor",
    "family",
    "platform",
    "model",
    "version",
    "loginIp",
    "snHw",
    "devType",
]
STR_OPERATORS = {
    "=": "eq",
    "!=": "neq",
    "i=": "ieq",
    "i!=": "nieq",
    "~": "like",
    "!": "notlike",
    "=~": "reg",
    "!=~": "nreg",
    "i=~": "ireg",
    "i!=~": "nireg",
}
INT_OPERATORS = {
    "=": "eq",
    "!=": "neq",
    ">": "gt",
    ">=": "gte",
    "<": "lt",
    "<=": "lte",
}


def main() -> dict[str, dict[str, Union[str, list]]]:
    arg_parser = argparse.ArgumentParser(
        description="Search the Inventory > Device table and return the results.",
        formatter_class=argparse.RawTextHelpFormatter,
        epilog=f"""
Valid Column Names:
{json.dumps(COLUMNS_HELP, indent=2)}

String Operators:
{json.dumps(STR_OPERATORS, indent=2)}

Number Operators:
{json.dumps(INT_OPERATORS, indent=2)}
""",
    )
    arg_parser.add_argument(
        "search",
        help="Search value: 'ipf_device_search rtr1'. Default uses 'hostname' for search. "
        "Example for different column: 'ipf_device_search vendor cisco'.",
        nargs="+",
    )
    arg_parser.add_argument(
        "-o",
        "--operator",
        help="Operator used in searching; default is 'like'.",
        default="like",
    )
    arg_parser.add_argument(
        "-C",
        "--csv",
        help="Export to CSV format.",
        action="store_true",
        default=False,
    )
    arg_parser.add_argument(
        "-a",
        "--add-columns",
        help="Comma separated list of column names to add to output.",
        default=None,
    )
    arg_parser.add_argument(
        "-r",
        "--remove-columns",
        help="Comma separated list of column names to remove from output.",
        default=None,
    )
    arg_parser.add_argument(
        "-A",
        "--attribute",
        help="Do a Device Attribute search (operator chosen is applied to both name and value): "
        "'ipf_device_search -A attribute_name attribute_value'.",
        action="store_true",
        default=False,
    )
    arg_parser.add_argument(
        "-B",
        "--box-style",
        help="Table box style (see https://rich.readthedocs.io/en/stable/appendix/box.html#appendix-box for options).",
        default="HEAVY_HEAD",
    )
    args = shared_args(arg_parser, "ipf_device_search")

    column, columns, search, operator = validate_args(args)
    columns = modify_columns(columns, args.add_columns, args.remove_columns)

    results, filters, url = query_ipf(args, column, search, operator, columns)
    if args.json and args.rich_disable:
        print(json.dumps(results, indent=2))
    elif args.json:
        CONSOLE.print(results)
    else:
        styles = [_ for _ in dir(box) if isinstance(getattr(box, _), box.Box)]
        # Export BOX_STYLE=MARKDOWN before running commands for docs.
        box_style = os.getenv("BOX_STYLE").upper() if os.getenv("BOX_STYLE", None) else args.box_style.upper()
        if box_style in styles:
            box_style = getattr(box, box_style)
        else:
            LOGGER.warning(
                f'Box style "{box_style}" is not valid, defaulting to `HEAVY_HEAD`. ' f"Valid styles: {styles}."
            )
            box_style = box.HEAVY_HEAD
        rich_print(results, filters, url, columns, args.count, box_style)


def query_ipf(args: argparse.Namespace, column: str, search: str, operator: str, columns: list):
    ipf = IPFClient(snapshot_id=args.snapshot, base_url=args.base_url, auth=args.auth)
    ipf._client.headers["user-agent"] += "; ipf_device_search"
    if args.attribute:
        results = attribute_search(ipf, column, search, operator, columns)
        columns.insert(0, "attributes")
        if args.csv:
            writer = csv.DictWriter(
                sys.stdout, fieldnames=columns, lineterminator="\n", quotechar='"', quoting=csv.QUOTE_ALL
            )
            writer.writeheader()
            writer.writerows(results)
            exit(0)
        return results, None, None
    else:
        filters = {column: [operator, search]}
        if args.csv:
            print(ipf.inventory.devices.all(export="csv", filters=filters, columns=columns).decode())
            exit(0)
        return (
            ipf.inventory.devices.all(filters=filters, columns=columns),
            filters,
            ipf.oas["tables/inventory/devices"].post.filter_url(filters, ipf._client.base_url),
        )


def attribute_search(ipf: IPFClient, name: str, value: str, operator: str, columns: list):
    # Search Snapshot Attributes table for matching attributes.
    snap_attr = ipf.fetch_all(
        "tables/snapshot-attributes",
        columns=["sn", "name", "value"],
        filters={"and": [{"name": [operator, name]}, {"value": [operator, value]}]},
    )
    filters, dev_attr = defaultdict(set), defaultdict(list)
    for _ in snap_attr:
        dev_attr[_["sn"]].append(_)
        filters[_["name"]].add(_["value"])
    cols = columns.copy()
    if "sn" not in cols:
        cols.append("sn")
    results = dict()
    for key, value in filters.items():
        f = {key: list(value)}
        LOGGER.info(f"Querying using Attribute Filter: {f}.")
        # Attribute filters are "and" not "or" so we must do multiple searches based on Attribute key.
        for data in ipf.inventory.devices.all(attr_filters=f, columns=cols):
            sn = data["sn"] if "sn" in columns else data.pop("sn")
            if sn in results:
                # Do not return duplicates if it matches multiple attributes.
                continue
            data.update({"attributes": ",".join([f"{_['name']}={_['value']}" for _ in dev_attr[sn]])})
            results[sn] = data
    return list(results.values())


def modify_columns(columns: list, add: str, remove: str):
    cols = set(columns)
    if add:
        for _ in add.split(","):
            col = validate_column(_.lower())
            if col and col not in cols:
                columns.append(col)
    if remove:
        for _ in remove.split(","):
            col = validate_column(_.lower())
            if col and col in cols:
                columns.remove(_)
    return columns


def validate_column(column: str) -> str:
    if not column:
        return column
    if column not in VALID_COLUMNS and column not in VALID_ALT_COLUMNS:
        LOGGER.critical(
            f"Column '{column}' is not a valid column name. "
            f"Valid names:\n{json.dumps(sorted(COLUMNS_HELP), indent=2)}"
        )
        exit(1)
    return VALID_COLUMNS[column]["name"] if column in VALID_COLUMNS else VALID_ALT_COLUMNS[column]["name"]


def validate_operator(operator: str, col_filter: str):
    if col_filter == "str":
        if operator in STR_OPERATORS:
            operator = STR_OPERATORS[operator]
        elif operator not in STR_OPERATORS.values():
            LOGGER.critical(
                f"Operator '{operator}' is not valid for text column. "
                f"Valid operators:\n{json.dumps(list(STR_OPERATORS), indent=2)}"
            )
            exit(1)
    else:
        if operator in INT_OPERATORS:
            operator = INT_OPERATORS[operator]
        elif operator not in INT_OPERATORS.values():
            LOGGER.critical(
                f"Operator '{operator}' is not a valid for number column. "
                f"Valid operators:\n{json.dumps(list(INT_OPERATORS), indent=2)}"
            )
            exit(1)
    return operator


def validate_args(args: argparse.Namespace):
    if len(args.search) != 2 and args.attribute:
        LOGGER.critical(
            "Two positional arguments required for Attribute searching: "
            "'ipf_device_search -A attribute_name attribute_value'."
        )
        exit(1)
    if len(args.search) == 1:
        column, search, col_filter = "hostname", args.search[0], "str"
    elif len(args.search) == 2:
        if args.attribute:
            return (
                args.search[0],
                DEFAULT_COLUMNS.copy(),
                args.search[1],
                validate_operator(args.operator.lower(), "str"),
            )
        column, search = args.search[0].lower(), args.search[1]
        column = validate_column(column)
        col_filter = VALID_COLUMNS[column.lower()]["filter"]
    else:
        LOGGER.critical("Too many positional arguments given.")
        exit(1)

    operator = validate_operator(args.operator.lower(), col_filter)

    columns = DEFAULT_COLUMNS.copy()
    if column in columns:
        columns.remove(column)
    columns.insert(0, column)
    if column in ["secDiscoveryDuration", "uptime"]:
        search = time_converter(search)
    return column, columns, search, operator


def time_converter(search: Union[str, int, float]) -> int:
    if isinstance(search, (int, float)):
        return int(search)
    time_duration_units = {
        "year": 31557600,
        "month": 2629800,
        "week": 604800,
        "day": 86400,
        "hour": 3600,
        "minute": 60,
        "second": 1,
    }
    time_convert = {
        **time_duration_units,
        "y": time_duration_units["year"],
        "mon": time_duration_units["month"],
        "w": time_duration_units["week"],
        "d": time_duration_units["day"],
        "h": time_duration_units["hour"],
        "min": time_duration_units["minute"],
        "m": time_duration_units["minute"],
        "sec": time_duration_units["second"],
        "s": time_duration_units["second"],
    }
    _ = re.findall(r"(\d*)\s?([a-z]*)", search.lower())

    seconds = 0
    for v, d in _:
        try:
            v = int(v)
        except ValueError:
            continue
        if d != "s" and d[-1] == "s":
            d = d[:-1]
        seconds += v * time_convert[d]

    return seconds


def rich_print(
    results: list, filters: dict, url: str, columns: list, count: bool = False, box_style: box.Box = box.HEAVY_HEAD
):
    if count or not results:
        CONSOLE.print(f"Total rows: {str(len(results))}")
    if not count and results:
        if filters:
            table = Table(*columns, title="Device Inventory", caption=f"Filter: {json.dumps(filters)}", box=box_style)
        else:
            table = Table(*columns, title="Device Inventory", box=box_style)
        for result in results:
            table.add_row(*[str(result[c]) for c in columns])
        CONSOLE.print(table)
    if url:
        CONSOLE.print(url, style=f"link {url}")


if __name__ == "__main__":
    main()



import logging
from datetime import datetime
from typing import Optional, Any, Union, ClassVar
from uuid import uuid4

from dateutil import parser
from pydantic import BaseModel, Field, PrivateAttr, field_validator, model_validator, model_serializer

from ipfabric.tools.shared import VALID_IP, validate_ip_network_str, raise_for_status

logger = logging.getLogger("ipfabric")


class Expiration(BaseModel):
    enabled: bool = False
    value: Optional[datetime] = None

    @model_validator(mode="after")
    def _verify(self):
        if self.enabled and not self.value:
            raise ValueError("Enabled Expiration requires a value.")
        elif not self.enabled and self.value:
            raise ValueError("Expiration value requires enabled to be True.")
        return self

    @model_serializer
    def _ser_model(self) -> dict[str, Any]:
        return (
            {"enabled": self.enabled, "value": self.value.strftime("%Y-%m-%d %H:%M:%S")}
            if self.enabled
            else {"enabled": self.enabled}
        )


class BaseCred(BaseModel):
    custom: bool = False
    username: str
    network: list[Union[str, VALID_IP]]  # TODO: Add v6 in 7.3
    excluded: Optional[list[Union[str, VALID_IP]]] = Field(default_factory=list, alias="excludeNetworks")
    expiration: Expiration = Field(None, alias="expirationDate")
    encrypt_password: str = Field(None, alias="password")
    priority: Optional[int] = None
    notes: Optional[str] = None

    @field_validator("network", "excluded")
    @classmethod
    def _verify_valid_networks(cls, v: list[str]) -> list[str]:
        return [validate_ip_network_str(_, ipv6=False) for _ in v]  # TODO: Add v6 in 7.3

    @model_serializer
    def _ser_model(self) -> dict[str, Any]:
        custom = {"custom": self.custom} if self.custom else dict()
        return dict(
            priority=self.priority,
            username=self.username,
            excludeNetworks=self.excluded,
            expirationDate=self.expiration.model_dump(),
            password=self.encrypt_password,
            notes=self.notes or "",
            **custom,
        )


class Credential(BaseCred, BaseModel):
    credential_id: Optional[str] = Field(None, alias="id")
    config_mgmt: bool = Field(False, alias="syslog")

    @model_validator(mode="after")
    def _verify(self):
        if not self.credential_id:
            self.custom = True
        return self

    @model_serializer
    def _ser_cred_model(self) -> dict[str, Any]:
        cred = self._ser_model()
        return dict(
            syslog=self.config_mgmt,
            id=self.credential_id if self.credential_id else str(uuid4()),
            network=self.network,
            **cred,
        )


class Privilege(BaseCred, BaseModel):
    privilege_id: str = Field(None, alias="id")
    network: list[Union[str, VALID_IP]] = Field(None, alias="includeNetworks")

    @model_validator(mode="after")
    def _verify(self):
        if not self.privilege_id:
            self.custom = True
        return self

    @model_serializer
    def _ser_cred_model(self) -> dict[str, Any]:
        cred = self._ser_model()
        return dict(id=self.privilege_id if self.privilege_id else str(uuid4()), includeNetworks=self.network, **cred)


def serialize_list(auth):
    if not auth:
        return list()
    priorities = sorted([_.priority for _ in auth])
    if priorities != list(range(1, len(priorities) + 1)):
        raise ValueError("Priorities are not sequential.")
    return [_.model_dump() for _ in auth]


class CredentialList(BaseModel):
    credentials: list[Credential]

    @model_serializer
    def _verify_priorities(self) -> dict[str, Any]:
        return serialize_list(self.credentials)

    def __bool__(self):
        return bool(self.credentials)


class PrivilegeList(BaseModel):
    privileges: list[Privilege]

    @model_serializer
    def _verify_priorities(self) -> dict[str, Any]:
        return serialize_list(self.privileges)


class Authentication(BaseModel):
    client: Any = Field(exclude=True)
    _credentials: dict[int, Credential] = PrivateAttr(default=dict())
    _enables: dict[int, Privilege] = PrivateAttr(default=dict())
    _cred_url: ClassVar[str] = "settings/credentials"
    _priv_url: ClassVar[str] = "settings/privileges"

    def model_post_init(self, __context: Any) -> None:
        self._credentials = self.get_credentials()
        self._enables = self.get_enables()

    @property
    def credentials(self) -> dict[int, Credential]:
        return self._credentials

    @property
    def enables(self) -> dict[int, Privilege]:
        return self._enables

    @property
    def credentials_by_id(self) -> dict[str, Credential]:
        return {_.credential_id: _ for _ in self._credentials.values()}

    @property
    def enables_by_id(self) -> dict[str, Privilege]:
        return {_.privilege_id: _ for _ in self._enables.values()}

    def get_credentials(self) -> dict[int, Credential]:
        """
        Get all credentials and sets them in the Authentication.credentials
        :return: self.credentials
        """
        res = raise_for_status(self.client.get(self._cred_url))
        self._credentials = {cred["priority"]: Credential(**cred) for cred in res.json()["data"]}
        return self._credentials

    def get_enables(self) -> dict[int, Privilege]:
        """
        Get all privileges (enable passwords) and sets them in the Authentication.enables
        :return:
        """
        res = raise_for_status(self.client.get(self._priv_url))
        self._enables = {priv["priority"]: Privilege(**priv) for priv in res.json()["data"]}
        return self._enables

    @staticmethod
    def _create_payload(username, password, notes, network, excluded, expiration):
        networks = network or ["0.0.0.0/0"]
        excluded = excluded or list()
        if expiration:
            expires = dict(
                enabled=True,
                value=parser.parse(expiration).strftime("%Y-%m-%d %H:%M:%S"),
            )
        else:
            expires = dict(enabled=False)
        payload = {
            "password": password,
            "username": username,
            "notes": notes or username,
            "excludeNetworks": [validate_ip_network_str(e, ipv6=False) for e in excluded],  # TODO: Add v6 in 7.3
            "expirationDate": expires,
            "network": [validate_ip_network_str(e, ipv6=False) for e in networks],
        }
        return payload

    def update_credential(
        self,
        credential: Credential,
        password: Optional[str] = None,
    ) -> Credential:
        """
        Updates a credential. Username cannot be changed, please delete and recreate.

        :param credential: Credential: Modified Credential object to update
        :param password: str: If updating the password then this is the unencrypted password

        :return: Credential: Obj: Credential Obj with ID and encrypted password
        """
        all_creds = {v.credential_id: v.model_dump(by_alias=True) for v in self.credentials.values()}
        if all_creds[credential.credential_id]["username"] != credential.username:
            raise SyntaxError("Cannot change username of credential, please delete and recreate.")
        all_creds[credential.credential_id] = credential.model_dump(by_alias=True)
        if password:
            all_creds[credential.credential_id].update({"password": password, "custom": True})
        raise_for_status(self.client.patch("settings", json={"credentials": list(all_creds.values())}))
        self.get_credentials()
        cred = self.credentials_by_id[credential.credential_id]
        logger.info(f"Updated credential with username {cred.username} and ID of {cred.credential_id}")
        return cred

    def create_credential(
        self,
        username: str,
        password: str,
        networks: list = None,
        notes: str = None,
        excluded: list = None,
        config_mgmt: bool = True,
        expiration: str = None,
    ) -> Credential:
        """
        Creates a new credential. Requires username and password and will default to all networks with no expiration.
        Does not default to use for configuration management, please set to true if needed.
        After creation Authentication.credentials will be updated with new priorities and the new cred is returned.
        :param username: str: Username
        :param password: str: Unencrypted password
        :param networks: list: List of networks defaults to ["0.0.0.0/0"]
        :param notes: str: Optional Note/Description of credential
        :param excluded: list: Optional list of networks to exclude
        :param config_mgmt: bool: Default True - do not use for configuration management
        :param expiration: str: Optional date for expiration, if none then do not expire.
                                To ensure correct date use YYYYMMDD or MM/DD/YYYY formats
        :return: Credential: Obj: Credential Obj with ID and encrypted password
        """
        payload = self._create_payload(username, password, notes, networks, excluded, expiration)
        payload.update({"syslog": config_mgmt})
        res = raise_for_status(self.client.post(self._cred_url, json=payload))
        self.get_credentials()
        cred = Credential(**res.json())
        logger.info(f"Created credential with username {cred.username} and ID of {cred.credential_id}")
        return cred

    def update_enable(
        self,
        enable: Privilege,
        password: Optional[str] = None,
    ) -> Credential:
        """
        Updates a enable privilege password. Username cannot be changed, please delete and recreate.

        :param enable: Privilege: Modified Credential object to update
        :param password: str: If updating the password then this is the unencrypted password

        :return: Privilege: Obj: Privilege Obj with ID and encrypted password
        """
        all_priv = {v.privilege_id: v.model_dump(by_alias=True) for v in self.enables.values()}
        if all_priv[enable.privilege_id]["username"] != enable.username:
            raise SyntaxError("Cannot change username of enable privilege, please delete and recreate.")
        all_priv[enable.privilege_id] = enable.model_dump(by_alias=True)
        if password:
            all_priv[enable.privilege_id].update({"password": password, "custom": True})
        raise_for_status(self.client.patch("settings", json={"privileges": list(all_priv.values())}))
        self.get_credentials()
        priv = self.enables_by_id[enable.privilege_id]
        logger.info(f"Updated enable privilege with username {priv.username} and ID of {priv.privilege_id}")
        return priv

    def create_enable(
        self,
        username: str,
        password: str,
        networks: list = None,
        notes: str = None,
        excluded: list = None,
        expiration: str = None,
    ) -> Privilege:
        """
        Creates a new enable password (privilege account).
        Requires username and password and will default to all networks with no expiration.
        After creation Authentication.enables will be updated with new priorities and the new enable is returned.
        :param username: str: Username
        :param password: str: Unencrypted password
        :param networks: list: List of networks defaults to ["0.0.0.0/0"]
        :param notes: str: Optional Note/Description of credential
        :param excluded: list: Optional list of networks to exclude
        :param expiration: str: Optional date for expiration, if none then do not expire.
                                To ensure correct date use YYYYMMDD or MM/DD/YYYY formats
        :return: Privilege: Obj: Privilege Obj with ID and encrypted password
        """
        payload = self._create_payload(username, password, notes, networks, excluded, expiration)
        payload["includeNetworks"] = payload.pop("network")
        res = raise_for_status(self.client.post(self._priv_url, json=payload))
        self.get_enables()
        priv = Privilege(**res.json())
        logger.info(f"Created enable password with username {priv.username} and ID of {priv.privilege_id}")
        return priv

    def delete_credential(self, credential: Union[Credential, str]) -> bool:
        """
        Deletes a credential and updates Authentication.credentials with new priorities.
        :param credential: Union[Credential, str]: Cred ID in a string or Credential object
        :return:
        """
        cred = credential.credential_id if isinstance(credential, Credential) else credential
        raise_for_status(self.client.request("DELETE", self._cred_url, json=[cred]))
        self.get_credentials()
        logger.warning(f"Deleted credential ID {cred}")
        return True

    def delete_enable(self, enable: Union[Privilege, str]) -> bool:
        """
        Deletes an enable password (privilege account) and updates Authentication.enable with new priorities.
        :param enable: Union[Privilege, str]: Enable ID in a string or Privilege object
        :return:
        """
        priv = enable.privilege_id if isinstance(enable, Privilege) else enable
        raise_for_status(self.client.request("DELETE", self._priv_url, json=[priv]))
        self.get_enables()
        logger.warning(f"Deleted enable password ID {priv}")
        return True

    def update_cred_priority(self, credentials: dict) -> dict:
        """
        Updates the priority of credentials.  Reorder Authentication.credentials dictionary and submit to this method.
        :param credentials: dict: {priority: Credential}
        :return: self.credentials: dict: {priority: Credential}
        """
        payload = [dict(id=c.credential_id, priority=p) for p, c in credentials.items()]
        raise_for_status(self.client.patch(self._cred_url, json=payload))
        self.get_credentials()
        return self.credentials

    def update_enable_priority(self, enables: dict) -> dict:
        """
        Updates the priority of enable passwords.  Reorder Authentication.enables dictionary and submit to this method.
        :param enables: dict: {priority: Privilege}
        :return: self.enables: dict: {priority: Privilege}
        """
        payload = [dict(id=e.privilege_id, priority=p) for p, e in enables.items()]
        raise_for_status(self.client.patch(self._priv_url, json=payload))
        self.get_enables()
        return self.enables



import logging
import re
import sys
from datetime import datetime, timezone
from http.cookiejar import CookieJar
from importlib import metadata
from ssl import SSLContext
from time import sleep
from typing import Optional, Union, Generator, Any, Type, Literal, Mapping

import jwt
from httpx import Client, BaseTransport, URL, Request, Response, Auth, Timeout, post, Proxy
from httpx._client import EventHook
from httpx._types import CertTypes, URLTypes
from packaging.version import Version
from pydantic import field_validator, Field, AliasChoices, PrivateAttr, InstanceOf
from pydantic_settings import BaseSettings, SettingsConfigDict, PydanticBaseSettingsSource, InitSettingsSource

from ipfabric.tools.shared import raise_for_status, valid_snapshot

logger = logging.getLogger("ipfabric")

RE_VERSION = re.compile(r"v?(\d(\.\d*)?)")
ProxyTypes = Union[None, URLTypes, InstanceOf[Proxy]]
TimeoutTypes = Union[
    Optional[float], tuple[Optional[float], Optional[float], Optional[float], Optional[float]], Timeout, None
]


def log_request(request: Request):
    logger.debug(f"Request event hook: {request.method} {request.url} - Waiting for response")


def log_response(response: Response):
    request = response.request
    logger.debug(f"Response event hook: {request.method} {request.url} - Status {response.status_code}")


class RateLimiter:
    def __init__(self, api_version: str = None):
        self.os_api_version = api_version

    def deprecated(self, response: Response):
        if response.headers.get("deprecation", False) == "true":
            version = response.url.path.split("/")[2]
            if version == self.os_api_version:
                logger.warning(
                    f"API endpoint '{response.url.path}' has deprecation header set and will be removed in a future release."
                )
            else:
                logger.info(
                    f"API endpoint '{response.url.path}' is using older API version, "
                    f"current IP Fabric version is {self.os_api_version}."
                )

    @staticmethod
    def rate_limit(response: Response, request: Request):
        reset = int(response.headers.get("X-RateLimit-Reset", 0))
        wait = reset - int(datetime.now(timezone.utc).timestamp()) + 0.5
        if wait > 0:
            logger.warning(f"Rate Limit Reached. Waiting for {wait} seconds.")
            sleep(wait)
        return request


class JWTToken(Auth, RateLimiter):
    """Used for streamlit."""

    def auth_flow(self, request: Request) -> Generator[Request, Response, None]:
        response = yield request
        self.deprecated(response)

        if response.status_code == 429 and response.headers.get("X-RateLimit-Remaining", None) == "0":
            yield self.rate_limit(response, request)
        if response.status_code == 401:
            logger.warning("Access Token has expired, please refresh IP Fabric or log in again.")
        return response


class TokenAuth(Auth, RateLimiter):
    def __init__(self, token: str, api_version: str = None) -> None:
        super().__init__(api_version)
        self._auth_header = token

    def auth_flow(self, request: Request) -> Generator[Request, Response, None]:
        request.headers["X-API-Token"] = self._auth_header

        response = yield request
        self.deprecated(response)
        if response.status_code == 429 and response.headers.get("X-RateLimit-Remaining", None) == "0":
            yield self.rate_limit(response, request)
        return response


class AccessToken(Auth, RateLimiter):
    def __init__(
        self,
        cookie_jar: CookieJar,
        username: str,
        password: str,
        base_url: URL,
        verify: Optional[Union[SSLContext, str, bool]],
        api_version: str = None,
    ):
        super().__init__(api_version)
        self.cookie_jar = cookie_jar
        self._login(username, password, base_url, verify)

    def auth_flow(self, request: Request) -> Generator[Request, Response, None]:
        response = yield request
        self.deprecated(response)

        if response.status_code == 429 and response.headers.get("X-RateLimit-Remaining", None) == "0":
            yield self.rate_limit(response, request)
        if response.status_code == 401:
            response.read()
            if "API_EXPIRED_ACCESS_TOKEN" in response.text:
                # Use refreshToken in Cookies to get new accessToken & Response updates accessToken in shared CookieJar
                resp = raise_for_status(post(response.url.join("/api/auth/token"), cookies=self.cookie_jar))
                request.headers["Cookie"] = "accessToken=" + resp.cookies["accessToken"]  # Update request
                yield request
        return response

    def _login(
        self, username: str, password: str, base_url: URL, verify: Optional[Union[SSLContext, str, bool]]
    ) -> None:
        raise_for_status(
            post(
                base_url.join("auth/login"),
                json=dict(username=username, password=password),
                cookies=self.cookie_jar,
                verify=verify,
            )
        )


class MyInitSettingsSource(InitSettingsSource):
    def __init__(self, settings_cls, init_kwargs: dict[str, Any]):
        timeout = init_kwargs.pop("timeout", "DEFAULT")
        init_kwargs = {k: v for k, v in init_kwargs.items() if v is not None}
        if timeout != "DEFAULT":
            init_kwargs["timeout"] = timeout
        super().__init__(settings_cls, init_kwargs)


class Setup(BaseSettings):
    model_config = SettingsConfigDict(env_file=".env", env_file_encoding="utf-8", env_prefix="ipf_", extra="allow")
    base_url: Union[str, URL] = Field(None, validation_alias=AliasChoices("base_url", "ipf_url"))
    api_version: Optional[Union[int, float, str]] = Field(
        None, validation_alias=AliasChoices("api_version", "ipf_version")
    )
    auth: Optional[Any] = Field(None, alias="auth", exclude=True)
    token: Optional[str] = None
    username: Optional[str] = None
    password: Optional[str] = None
    snapshot_id: Union[str, None] = Field("$last", validation_alias=AliasChoices("snapshot_id", "ipf_snapshot"))
    verify: Union[SSLContext, bool, str] = True
    timeout: Union[TimeoutTypes, Literal["DEFAULT"]] = Timeout(timeout=5.0)
    nvd_api_key: Optional[str] = Field(None, alias="nvd_api_key")
    debug: bool = False
    http2: bool = True
    proxy: Optional[Union[ProxyTypes, None]] = None
    mounts: Optional[Mapping[str, Optional[BaseTransport]]] = None
    cert: Optional[Union[CertTypes, None]] = None
    event_hooks: Optional[Mapping[str, list[EventHook]]] = None
    _cookie_jar: CookieJar = PrivateAttr(default_factory=CookieJar)
    client: Optional[Client] = Field(None, exclude=True)
    os_version: Optional[str] = Field(None)
    _os_api_version: Optional[str] = PrivateAttr(None)
    _auth_type: Optional[str] = PrivateAttr(None)

    @classmethod
    def settings_customise_sources(
        cls,
        settings_cls: Type[BaseSettings],
        init_settings: PydanticBaseSettingsSource,
        env_settings: PydanticBaseSettingsSource,
        dotenv_settings: PydanticBaseSettingsSource,
        file_secret_settings: PydanticBaseSettingsSource,
    ) -> tuple[PydanticBaseSettingsSource, ...]:
        return (
            MyInitSettingsSource(settings_cls, init_settings.init_kwargs),
            env_settings,
            dotenv_settings,
            file_secret_settings,
        )

    def model_post_init(self, __context):
        event_hooks = {"request": [log_request], "response": [log_response]} if self.debug else self.event_hooks
        self.client = Client(
            base_url=self.base_url,
            cookies=self._cookie_jar,
            http2=self.http2,
            verify=self.verify,
            timeout=self.timeout,
            headers={
                "User-Agent": f'python-ipfabric-sdk/{metadata.version("ipfabric")} (Python {sys.version.split(" ")[0]})'
            },
            event_hooks=event_hooks,
            proxy=self.proxy,
            mounts=self.mounts,
            cert=self.cert,
        )
        if not (self.auth or self.token) and not (self.username and self.password):
            raise RuntimeError("IP Fabric Authentication not provided.")
        self.api_version, self.os_version, self._os_api_version = self.check_version(self.api_version)
        self.base_url = self.base_url.join(f"api/{self.api_version}/")
        self.client.base_url = self.base_url
        self._verify_auth()

    def _verify_auth(self):
        if self.auth:
            self.client.auth = self._get_auth_from_auth()
        elif self.token:
            self.client.auth = self._check_jwt(self.token)
        elif self.username and self.password:
            self.client.auth = self._get_access_token_auth(self.username, self.password)

    def _check_jwt(self, token):
        try:
            jwt.decode(token, options={"verify_signature": False})
            self.client.cookies.set("accessToken", token)
            self._auth_type = "JWT_AUTH"
            return JWTToken(api_version=self._os_api_version)
        except jwt.exceptions.DecodeError:
            self._auth_type = "API_TOKEN_AUTH"
            return TokenAuth(token, api_version=self._os_api_version)

    def _get_auth_from_auth(self) -> Auth:
        """Separate auth handling logic"""
        if isinstance(self.auth, str):
            return self._check_jwt(self.auth)
        elif isinstance(self.auth, tuple):
            return self._get_access_token_auth(self.auth[0], self.auth[1])
        self._auth_type = type(self.client.auth)
        return self.auth

    def _get_access_token_auth(self, username: str, password: str) -> Auth:
        """Separate auth handling logic"""
        self._auth_type = "USER_PASS_ACCESS_TOKEN"
        return AccessToken(
            self._cookie_jar,
            username,
            password,
            self.base_url,
            self.verify,
            api_version=self._os_api_version,
        )

    @property
    def update_attrs(self) -> dict:
        return {
            "base_url": self.base_url,
            "api_version": self.api_version,
            "auth": self._auth_type,
            "timeout": self.client.timeout,
            "_os_version": self.os_version,
            "_os_api_version": self._os_api_version,
            "debug": self.debug,
            "_client": self.client,
            "verify": self.verify,
            "proxy": self.proxy,
            "mounts": self.mounts,
            "cert": self.cert,
            "event_hooks": self.event_hooks,
            "http2": self.http2,
            "nvd_api_key": self.nvd_api_key,
        }

    def check_version(self, custom_api_version) -> tuple:
        """Checks API Version and returns the version to use in the URL and the OS Version

        Returns:
            api_version, os_version
        """
        cfg_version = Version(custom_api_version)
        api_version = f"v{cfg_version.major}.{cfg_version.minor}"

        resp = raise_for_status(self.client.get("/api/version")).json()
        os_api_version = Version(resp["apiVersion"])

        if os_api_version.major != cfg_version.major:
            raise RuntimeError(
                f"OS Major Version `{os_api_version.major}` does not match SDK Major Version `{cfg_version.major}`."
            )

        if cfg_version.minor > os_api_version.minor:
            logger.warning(
                f"Specified SDK Version `{api_version}` is greater than "
                f"OS API Version `{os_api_version}`, using OS Version."
            )
            api_version = f"v{os_api_version.base_version}"

        return api_version, resp["releaseVersion"], resp["apiVersion"]

    @field_validator("api_version")
    @classmethod
    def _valid_version(cls, v: Union[None, int, float, str]) -> Union[None, str]:
        if not v:
            return metadata.version("ipfabric")
        re_version = RE_VERSION.match(str(v))
        if not re_version:
            raise ValueError(f"IPF_VERSION ({v}) is not valid, must be like `v#` or `v#.#`.")
        elif re_version and re_version.group(2):
            return "v" + RE_VERSION.match(str(v)).group(1)
        else:
            return "v" + RE_VERSION.match(str(v)).group(1) + ".0"

    @field_validator("base_url")
    @classmethod
    def _convert_url(cls, v: Union[URL, str]) -> URL:
        if isinstance(v, str):
            v = URL(v)
        return v

    @field_validator("snapshot_id")
    @classmethod
    def _valid_snapshot(cls, v: Union[None, str]) -> str:
        return valid_snapshot(v, init=True)

    @field_validator("verify")
    @classmethod
    def _verify(cls, v: Union[bool, int, str]) -> Union[bool, str]:
        if isinstance(v, bool):
            return v
        elif isinstance(v, int):
            return bool(v)
        elif v.lower() in {"0", "off", "f", "false", "n", "no", "1", "on", "t", "true", "y", "yes"}:
            return False if v.lower() in {0, "0", "off", "f", "false", "n", "no"} else True
        else:
            return v

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Needed for context"""
        pass



from typing import Union, Any

from pydantic import BaseModel, Field, AliasChoices, ConfigDict, AliasGenerator

from ipfabric.models.jobs import Job


class ExportConfig(BaseModel):
    model_config = ConfigDict(
        alias_generator=AliasGenerator(
            serialization_alias=lambda field_name: field_name.replace("_", "-"),
        )
    )
    device_attributes: bool = Field(True, validation_alias=AliasChoices("device-attributes", "device_attributes"))
    jumphost: bool = True
    oui: bool = True
    discovery: bool = True
    discovery_seeds: bool = Field(True, validation_alias=AliasChoices("discovery-seeds", "discovery_seeds"))
    device_credentials: bool = Field(True, validation_alias=AliasChoices("device-credentials", "device_credentials"))
    assurance_engine: bool = Field(True, validation_alias=AliasChoices("assurance-engine", "assurance_engine"))
    discovery_tasks: bool = Field(True, validation_alias=AliasChoices("discovery-tasks", "discovery_tasks"))
    vendors_api: bool = Field(True, validation_alias=AliasChoices("vendors-api", "vendors_api"))
    site_separation: bool = Field(True, validation_alias=AliasChoices("site-separation", "site_separation"))
    cli_settings: bool = Field(True, validation_alias=AliasChoices("cli-settings", "cli_settings"))
    bgp_routing: bool = Field(True, validation_alias=AliasChoices("bgp-routing", "bgp_routing"))
    snapshot_retention: bool = Field(True, validation_alias=AliasChoices("snapshot-retention", "snapshot_retention"))
    configuration_management: bool = Field(
        True, validation_alias=AliasChoices("configuration-management", "configuration_management")
    )
    api_tokens: bool = Field(True, validation_alias=AliasChoices("api-tokens", "api_tokens"))
    webhooks: bool = True
    snmp: bool = True
    certificates: bool = True
    local_users: bool = Field(True, validation_alias=AliasChoices("local-users", "local_users"))
    ldap: bool = True
    roles: bool = True
    policies: bool = True
    intent_checks: bool = Field(True, validation_alias=AliasChoices("intent_checks", "intent_checks"))
    dashboards_and_widgets: bool = Field(
        True, validation_alias=AliasChoices("dashboards-and-widgets", "dashboards_and_widgets")
    )

    def export(self) -> dict[str, Any]:
        return {
            "configurationOptions": [k for k, v in self.model_dump(by_alias=True).items() if v is True],
        }


class ImportJob(BaseModel):
    job: Job

    @property
    def job_completed(self):
        return self.job.isDone is True and self.job.status == "done"


class ExportJob(ImportJob, BaseModel):
    config: Union[dict, None] = None



from typing import Any, Union

from httpx import Client, Timeout
from pydantic import BaseModel, PrivateAttr

from ipfabric.models.matrix import Version, SupportMatrix


class FeatureMatrix(BaseModel):
    timeout: Any = Timeout(timeout=5.0)
    proxy: Any = None
    _client: Client = PrivateAttr(None)
    _versions: list[Version] = PrivateAttr(None)
    _matrix_version: Version = PrivateAttr(None)
    _matrix: Any = PrivateAttr(None)

    def model_post_init(self, __context: Any) -> None:
        self._client = Client(base_url="https://matrix.ipfabric.io/api/", timeout=self.timeout, proxy=self.proxy)
        self._versions = [Version(**_) for _ in self._client.get("version").json()]
        self._versions.sort(key=lambda p: [int(_) for _ in p.name.split(".")], reverse=True)
        self.create_matrix()

    @property
    def versions(self) -> list[Version]:
        """Descending Ordered List of Versions; Newest is index 0."""
        return self._versions

    @property
    def matrix_version(self) -> Version:
        return self._matrix_version

    @property
    def matrix(self) -> SupportMatrix:
        return self._matrix

    def create_matrix(self, version: Union[Version, str] = None) -> SupportMatrix:
        """Creates a matrix based on the version

        Args:
            version: Version object or string, defaults to latest

        Returns:
        """
        if isinstance(version, str):
            version = {_.name: _ for _ in self._versions}[version]
        elif version is None:
            version = self.versions[0]
        if version == self._matrix_version:
            return self._matrix
        vid = version.id
        categories = {_["id"]: _ for _ in self._client.get("task/category", params={"versionId": vid}).json()}
        tasks = {_["id"]: _ for _ in self._client.get("task", params={"versionId": vid}).json()}
        vendors = {_["id"]: _ for _ in self._client.get("vendor", params={"versionId": vid}).json()}
        families = {_["id"]: _ for _ in self._client.get("vendor/family", params={"versionId": vid}).json()}

        matrix = [
            dict(
                category=categories[tasks[status["taskId"]]["categoryId"]],
                task=tasks[status["taskId"]],
                vendor=vendors[status["vendorId"]],
                family=families[status["familyId"]],
                status=status,
            )
            for status in self._client.get("task/status", params={"versionId": vid}).json()
        ]
        matrix = SupportMatrix(matrix=matrix, version=version)

        self._matrix, self._matrix_version = matrix, version
        return self._matrix



"""
Python3 script to search for an IPv4, IPv6, or MAC address in multiple IP Fabric's tables and prints the output.
"""

import argparse
import json
import logging
from typing import Union

from ipfabric import IPFClient
from ipfabric.models.global_search import RouteTableSearch
from ipfabric.scripts.shared import shared_args

try:
    from rich.console import Console

    CONSOLE = Console()
except ImportError:
    CONSOLE = None

logging.basicConfig(format="%(levelname)s: %(message)s")

LOGGER = logging.getLogger("ipf_route_search")


def main() -> dict[str, dict[str, Union[str, list]]]:
    arg_parser = argparse.ArgumentParser(
        description="Search IPv4 or IPv6 Route tables for an Address or Subnet.",
        formatter_class=argparse.RawTextHelpFormatter,
        epilog="""
IPv4 and IPv6 Route or Next Hop IP:
    `=|eq`: Route (CIDR) or Next Hop IP exactly matches.
    `IP|ip`: IP is contained within Routed subnet or Next Hop IP matches.

IPv4 and IPv6 Next Hop IP:
    `cidr|CIDR`: Next Hop IP is contained within subnet.

IPv4 Route CIDR Searching:
    `>|gt`: CIDR is strict supernet (10.0.0.0/24 -> 10.0.0.0/[8..23], 0.0.0.0/0)
    `>=|gte`: CIDR is supernet (10.0.0.0/24 -> 10.0.0.0/[8..24], 0.0.0.0/0)
    `<|lt`: CIDR is strict subnet (10.0.0.0/24 -> 10.0.0.0/[25..32])
    `<|lte`: CIDR is subnet (10.0.0.0/24 -> 10.0.0.0/[24..32])
    `@|sect`: CIDR overlaps (10.0.0.0/24 -> 10.0.0.0/[8..32], 0.0.0.0/0)
    `!@|nsect`: CIDR does not overlap (10.0.0.0/24 -> 10.1.0.0/24)
""",
    )
    arg_parser.add_argument(
        "address",
        help="IPv4 or IPv6 Address/CIDR to search for.",
    )
    arg_parser.add_argument(
        "-o",
        "--operator",
        help="Default 'eq'",
        default="eq",
    )
    arg_parser.add_argument(
        "-d",
        "--default-route",
        help='Include "0.0.0.0/0" routes.',
        action="store_true",
        default=False,
    )
    arg_parser.add_argument(
        "-n",
        "--next-hop",
        help="Search Next Hop IP instead of Route.",
        action="store_true",
        default=False,
    )
    args = shared_args(arg_parser, "ipf_route_search")

    ipf = IPFClient(snapshot_id=args.snapshot, base_url=args.base_url, auth=args.auth)
    ipf._client.headers["user-agent"] += "; ipf_route_search"
    rs = RouteTableSearch(client=ipf)
    print_results(rs.search(args.address, args.operator, args.next_hop, not args.default_route), args)


def rich_print(result: dict, count: bool = False):
    CONSOLE.print(f"\n{result['menu']}: {str(len(result['data']))}" if count else f"\n{result['menu']}")
    CONSOLE.print(result["url"], style=f"link {result['url']}")
    if not count:
        json_str = json.dumps(result["data"], indent=4)
        CONSOLE.print(json_str)


def py_print(result: dict, count: bool = False):
    print(f"\n{result['menu']}: {str(len(result['data']))}" if count else f"\n{result['menu']}")
    print(result["url"])
    if not count:
        print(json.dumps(result["data"], indent=4))


def print_results(results: dict, args: argparse.Namespace):
    global CONSOLE
    if args.rich_disable:
        CONSOLE = None
    if not results["data"]:
        msg = "\nNo routes found."
        CONSOLE.print(msg) if CONSOLE else LOGGER.error(msg)
        exit(1)
    if args.json:
        CONSOLE.print(json.dumps(results, indent=4)) if CONSOLE else print(json.dumps(results, indent=4))
    elif CONSOLE:
        rich_print(results, args.count)
    else:
        py_print(results, args.count)


if __name__ == "__main__":
    main()



from .constants import VALID_NET_PROTOCOLS, VALID_PATH_PROTOCOLS
from .graph_parameters import (
    Unicast,
    Multicast,
    Host2GW,
    Network,
    OtherOptions,
    Algorithm,
    EntryPoint,
    Layout,
    ICMP,
)
from .graph_settings import (
    GroupSettings,
    EdgeSettings,
    NetworkSettings,
    PathLookupSettings,
    Overlay,
    Style,
    PathLookup,
)
from .shared_view import SharedView

__all__ = [
    "Unicast",
    "Multicast",
    "Host2GW",
    "Network",
    "OtherOptions",
    "Algorithm",
    "Overlay",
    "NetworkSettings",
    "PathLookupSettings",
    "EntryPoint",
    "VALID_NET_PROTOCOLS",
    "VALID_PATH_PROTOCOLS",
    "Layout",
    "SharedView",
    "GroupSettings",
    "EdgeSettings",
    "ICMP",
    "Style",
    "PathLookup",
]



import json
from collections import defaultdict
from functools import cached_property
from importlib.resources import files
from typing import Optional, Any, Union
from urllib.parse import quote_plus

from httpx import URL
from pydantic import BaseModel, PrivateAttr, TypeAdapter, Field, FilePath

from ipfabric.tools.shared import raise_for_status

OAS_DIR = files("ipfabric.oas")
CONTENT_TYPE = "application/json"


class NestedColumn(BaseModel):
    parent: str
    child: str


class Column(BaseModel):
    name: str
    filter: Optional[str] = None
    deprecated: Optional[bool] = False


class Scan(BaseModel):
    columns: Optional[list[Union[Column, NestedColumn]]] = Field(default_factory=list)
    full_scan: bool = True


class ComplexColumn(BaseModel):
    filter: Optional[str] = None
    array: Optional[bool] = Field(False)
    children: list[Column] = Field(default_factory=list)

    @property
    def children_by_name(self) -> dict[str, Column]:
        return {_.name: _ for _ in self.children}

    @property
    def children_by_filters(self) -> dict[str, list[Column]]:
        tmp = defaultdict(list)
        [tmp[_.filter].append(_) for _ in self.children]
        return dict(tmp)


class Endpoint(BaseModel):
    full_api_endpoint: str
    web_endpoint: Optional[str] = None
    columns: Optional[list[str]] = None
    nested_columns: Optional[dict[str, ComplexColumn]] = Field(default_factory=dict)
    array_columns: Optional[dict[str, ComplexColumn]] = Field(default_factory=dict)
    deprecated_columns: Optional[set[str]] = Field(default_factory=set)
    summary: Optional[str] = None
    title: Optional[str] = None
    tags: Optional[list[str]] = Field(default_factory=list)
    tag_groups: Optional[list[str]] = Field(default_factory=list)
    method: str
    ui_columns: Optional[list[str]] = None
    api_scope_id: Optional[str] = None
    description: Optional[str] = None
    ipv4: Optional[Scan] = Field(default_factory=Scan)
    ipv6: Optional[Scan] = Field(default_factory=Scan)
    mac: Optional[Scan] = Field(default_factory=Scan)
    deprecated: Optional[bool] = False

    def __repr__(self):
        return f"({self.method}, {self.api_endpoint})"

    @property
    def api_endpoint(self):
        return self.full_api_endpoint[1:]

    @property
    def web_menu(self):
        return f"{self.tag_groups[0]} — {self.tags[0]}" if self.tag_groups and self.tags else None

    def filter_url(self, filters: Union[str, dict], base_url: Union[URL, str]) -> Union[None, str]:
        if not self.web_endpoint:
            return None
        filters = json.loads(filters) if isinstance(filters, str) else filters
        filters = filters if "filters" in filters else {"filters": filters}
        url = base_url.join(self.web_endpoint) if isinstance(base_url, URL) else URL(base_url).join(self.web_endpoint)
        return f"{url}?options=" + quote_plus(json.dumps(filters, separators=(",", ":")))


class Methods(BaseModel):
    full_api_endpoint: str
    get: Optional[Endpoint] = None
    put: Optional[Endpoint] = None
    patch: Optional[Endpoint] = None
    post: Optional[Endpoint] = None
    delete: Optional[Endpoint] = None

    @property
    def api_endpoint(self):
        return self.full_api_endpoint[1:]


class OAS(BaseModel):
    client: Any = Field(exclude=True)
    local_oas: bool = True
    local_oas_file: Optional[Union[FilePath, dict]] = None
    _oas: dict[str, Methods] = PrivateAttr(default_factory=dict)

    def model_post_init(self, __context) -> None:
        self._oas = self._get_oas()

    @property
    def oas(self) -> dict[str, Methods]:
        return self._oas

    def _get_oas(self) -> dict[str, Methods]:
        if not self.local_oas or (self.local_oas_file and self.local_oas):
            return self._parse_oas()
        try:
            min_oas = OAS_DIR.joinpath(self.client.api_version + ".json").read_text(encoding="UTF-8")
            oas = TypeAdapter(dict[str, Methods]).validate_json(min_oas)
            return oas
        except FileNotFoundError:
            return self._parse_oas()

    @cached_property
    def web_to_api(self) -> dict[str, Endpoint]:
        return {m.post.web_endpoint: m.post for m in self._oas.values() if m.post and m.post.web_endpoint}

    @cached_property
    def scope_to_api(self) -> dict[str, Endpoint]:
        _ = dict()
        for methods in self._oas.values():
            for method in ["get", "put", "post", "patch", "delete"]:
                m = getattr(methods, method, None)
                if m and m.api_scope_id:
                    _[m.api_scope_id] = m
        return _

    def _complex_columns(self, data: Endpoint, spec: dict) -> Endpoint:
        x_columns = {_["key"]: _ for _ in spec.get("x-table", {}).get("columns", [])}

        try:
            columns = spec["responses"]["200"]["content"][CONTENT_TYPE]["schema"]["properties"]["data"]["items"][
                "properties"
            ]
            for k, v in columns.items():
                col_type = v.get("type", None)
                children = list()
                if k in x_columns and isinstance(x_columns[k].get("filter", None), dict):
                    children = [
                        Column(
                            name=c["field"],
                            filter=c["filter"]["type"] if isinstance(c["filter"], dict) else c["filter"],
                            deprecated=c.get("deprecated", False),
                        )
                        for c in x_columns[k]["filter"]["children"]
                    ]
                if col_type == "array":
                    filters = v["filter"] if isinstance(v.get("filter", None), str) else None
                    data.array_columns[k] = ComplexColumn(array=True, children=children, filter=filters)
                    if v.get("items", {}).get("type", None) == "object":
                        data.nested_columns[k] = data.array_columns[k]
                elif col_type == "object":
                    data.nested_columns[k] = ComplexColumn(children=children)
        except KeyError:
            pass
        return self._global_search(data, spec)

    def _post_logic(self, data: Endpoint, spec: dict):
        data.web_endpoint = spec.get("x-table", {}).get("webPath")
        data.title = spec.get("x-table", {}).get("title")
        data.ui_columns = [_["key"] for _ in spec.get("x-table", {}).get("columns", [])]
        columns = set(
            spec.get("requestBody", {})
            .get("content", {})
            .get(CONTENT_TYPE, {})
            .get("schema", {})
            .get("properties", {})
            .get("columns", {})
            .get("items", {})
            .get("enum", [])
        )
        data.columns = list(columns)
        return self._complex_columns(data, spec)

    def _complex_global_search(self, data: Endpoint) -> Endpoint:
        for cname, column in data.nested_columns.items():
            data.ipv4.columns.extend(
                [NestedColumn(parent=cname, child=_.name) for _ in column.children_by_filters.get("ip", [])]
            )
            data.ipv6.columns.extend(
                [NestedColumn(parent=cname, child=_.name) for _ in column.children_by_filters.get("ipv6", [])]
            )
            data.mac.columns.extend(
                [
                    NestedColumn(parent=cname, child=_.name)
                    for _ in column.children
                    if self._check_mac(_.name, data.full_api_endpoint)
                ]
            )
        return data

    @staticmethod
    def _check_mac(mac: str, endpoint: str) -> bool:
        return (
            "mac" in mac
            and mac not in ["macVerification", "macFlags", "macCount"]
            and endpoint not in ["/tables/inventory/devices"]
        )

    def _global_search(self, data: Endpoint, spec: dict) -> Endpoint:
        if not data.api_endpoint.startswith("tables"):
            return data
        columns = spec.get("x-table", {}).get("columns", [])
        for column in columns:
            key, filter_type = column["key"], column.get("filter", "text")
            if (
                filter_type in ["ip", "routing"]
                and (key != "loginIp" or (key == "loginIp" and data.full_api_endpoint == "/tables/inventory/devices"))
                and data.full_api_endpoint not in ["/tables/routing/protocols/ospf-v3/neighbors"]
            ):
                data.ipv4.columns.append(Column(name=key, filter=filter_type))
            elif filter_type in ["ipv6", "routingIpv6"]:
                data.ipv6.columns.append(Column(name=key, filter=filter_type))
            elif self._check_mac(key, data.full_api_endpoint):
                data.mac.columns.append(Column(name=key, filter=filter_type))
        return self._complex_global_search(data)

    def _parse_oas(self) -> dict[str, Methods]:
        if not self.local_oas or not self.local_oas_file:
            url = self.client.base_url.join("/api/static/oas/openapi-internal.json")
            oas = raise_for_status(self.client.get(url, follow_redirects=True)).json()
        elif isinstance(self.local_oas_file, dict):
            oas = self.local_oas_file
        else:
            with open(self.local_oas_file, "r") as f:
                oas = json.load(f)

        endpoints = dict()
        for endpoint, methods in oas["paths"].items():
            methods_obj = Methods(full_api_endpoint=endpoint)
            for method, spec in methods.items():
                data = Endpoint(
                    full_api_endpoint=endpoint,
                    method=method,
                    api_scope_id=spec.get("x-apiScopeId", None),
                    summary=spec.get("summary", None),
                    description=spec.get("description", None),
                    tags=spec.get("tags", None),
                    tag_groups=spec.get("x-tagGroups", None),
                    deprecated=spec.get("deprecated", False),
                    deprecated_columns={
                        _["key"] for _ in spec.get("x-table", {}).get("columns", []) if _.get("deprecated", False)
                    },
                )
                if method == "post":
                    data = self._post_logic(data, spec)
                setattr(methods_obj, method, data)
            endpoints[endpoint[1:]] = methods_obj
        return endpoints



import argparse
import logging
import os

from ipfabric import IPFClient
from ipfabric.models.jobs import TechsupportPayload, Jobs, TechsupportSnapshotSettings
from ipfabric.scripts.shared import load_env, base_args, parse_base_args

try:
    from rich.console import Console
    from rich.progress import Progress, SpinnerColumn, TextColumn
except ImportError:
    raise ImportError("rich is required, please install by using `pip install rich or pip install ipfabric[cli]`.")


logger = logging.getLogger("TechSupportCLI")

PROGRESS = "[progress.description]{task.description}"
console = Console()


def parse_args(parser: argparse.ArgumentParser) -> argparse.Namespace:
    args = parse_base_args(parser.parse_args())

    if not args.upload_username and not args.upload_password:
        console.log(args)
        raise ValueError(
            "Upload username and password must be provided as an argument or set as an environment variable "
            "(IPF_UPLOAD_USERNAME and IPF_UPLOAD_PASSWORD)."
        )
    return args


def main():  # noqa:C901
    load_env()
    parser = argparse.ArgumentParser(
        description="Upload a techsupport file using IPFabric API. "
        "Environment variables can be exported or configured in the .env file."
    )
    parser = base_args(parser)
    parser.add_argument(
        "-uu",
        "--upload-username",
        default=os.getenv("IPF_UPLOAD_USERNAME", "techsupport"),
        help="Username for uploading the techsupport file or IPF_UPLOAD_USERNAME env variable. Default: techsupport",
    )
    parser.add_argument(
        "-up",
        "--upload-password",
        default=os.getenv("IPF_UPLOAD_PASSWORD", None),
        help="Password for uploading the techsupport file or IPF_UPLOAD_PASSWORD env variable.",
    )
    parser.add_argument(
        "-ur",
        "--upload-region",
        choices=["eu", "us"],
        default=os.getenv("IPF_UPLOAD_REGION", "eu"),
        help="Server region for the upload or IPF_UPLOAD_REGION env variable. Default: eu",
    )
    parser.add_argument(
        "-ut",
        "--upload-timeout",
        type=int,
        default=600,
        help="Timeout for the upload request in seconds. Default: 600",
    )
    parser.add_argument(
        "-nvu",
        "--no-verify-upload",
        dest="upload_verify",
        action="store_false",
        default=True,
        help="Disable verification of the upload process.",
    )
    parser.add_argument(
        "-dt",
        "--download-timeout",
        type=int,
        default=60,
        help="Timeout to wait for the download process in seconds. Default: 60",
    )
    parser.add_argument(
        "-dr",
        "--download-retry",
        type=int,
        default=5,
        help="How many times to wait --download-timeout before failing Default: 5 "
        "(5 Retries of 60 seconds each equals 5 minutes.)",
    )
    parser.add_argument(
        "-DR",
        "--dry-run",
        action="store_true",
        help="Dry run the techsupport process without actually uploading the file.",
    )
    args = parse_args(parser)

    console.log(f"[bold blue]Initializing Techsupport API with base_url={args.base_url}...[/bold blue]")
    ipf = IPFClient(snapshot_id=args.snapshot, base_url=args.base_url, auth=args.auth)
    ipf._client.headers["user-agent"] += "; ipf_techsupport"
    jobs = Jobs(client=ipf)

    payload = TechsupportPayload(snapshot=TechsupportSnapshotSettings(id=args.snapshot))
    if args.dry_run:
        console.log("[bold yellow]Dry run enabled. Skipping techsupport job creation...[/bold yellow]")
        return
    console.log("[bold blue]Starting techsupport job...[/bold blue]")

    with Progress(
        SpinnerColumn(),
        TextColumn(PROGRESS),
        console=console,
    ) as progress:
        _ = progress.add_task(description="Generating techsupport...", total=None)
        try:
            job = jobs.generate_techsupport(
                payload, wait_for_ts=True, timeout=args.download_timeout, retry=args.download_retry
            )
        except Exception as e:
            progress.stop()
            console.log(f"[bold red]Failed to generate techsupport: {e}[/bold red]")
            return

    console.log("[bold blue]Downloading techsupport file...[/bold blue]")
    with Progress(
        SpinnerColumn(),
        TextColumn("[progress.description]{task.description}"),
        console=console,
    ) as progress:
        task = progress.add_task(description="Downloading techsupport file...", total=None)
        try:
            techsupport_resp = jobs.download_techsupport_file(job.id)
            progress.update(task, description="Techsupport file downloaded successfully.")
        except Exception as e:
            progress.stop()
            console.log(f"[bold red]Failed to download techsupport file: {e}[/bold red]")
            return

    console.log("[bold blue]Uploading techsupport file...[/bold blue]")
    with Progress(
        SpinnerColumn(),
        TextColumn(PROGRESS),
        console=console,
    ) as progress:
        task = progress.add_task(description="Uploading file...", total=None)
        try:
            jobs.upload_techsupport_file(
                upload_username=args.upload_username,
                upload_password=args.upload_password,
                upload_file_timeout=args.upload_timeout,
                upload_server=args.upload_region,
                techsupport_bytes=techsupport_resp.content,
                upload_verify=args.upload_verify,
            )
            progress.update(task, description="Techsupport file uploaded successfully.")
        except Exception as e:
            progress.stop()
            console.log(f"[bold red]Failed to upload techsupport file: {e}[/bold red]")
            return

    console.log("[bold green]Techsupport process completed successfully.[/bold green]")


if __name__ == "__main__":
    main()



import logging
import re
from datetime import datetime
from typing import Optional, Any

from pydantic import Field, BaseModel

from ipfabric.tools.shared import raise_for_status

logger = logging.getLogger("ipfabric")

ATTR_REGEX = re.compile(r"^[a-zA-Z]\w*[a-zA-Z0-9]+$")


class Attributes(BaseModel):
    """
    Class to retrieve Global and Local (Snapshot specific) Attributes.  You must specify snapshot_id to use local.
    """

    client: Any = Field(description="IPFClient", exclude=True)
    snapshot_id: Optional[str] = Field(default=None, description="Snapshot ID to switch to Local Attributes")

    def model_post_init(self, __context: Any) -> None:
        if isinstance(self.snapshot_id, str):
            self.snapshot_id = self.client.get_snapshot(self.snapshot_id).snapshot_id

    @property
    def endpoint(self):
        return "attributes/local" if self.snapshot_id else "attributes/global"

    @property
    def post_endpoint(self):
        return "tables/snapshot-attributes" if self.snapshot_id else "tables/global-attributes"

    @staticmethod
    def check_attribute_name(attributes: set):
        invalid = list()
        for attribute in attributes:
            if not ATTR_REGEX.match(attribute):
                invalid.append(attribute)
        if invalid:
            raise NameError(
                f"The following Attribute Names are invalid and do match regex rule "
                f'"^[a-zA-Z][a-zA-Z0-9_]*[a-zA-Z0-9]+$":\n{invalid}'
            )
        return True

    def all(
        self,
        columns: list = None,
        filters: Optional[dict] = None,
        sort: Optional[dict] = None,
    ):
        """Gets all data from corresponding endpoint

        Args:
            columns: Optional columns to return, default is all
            filters: Optional filters
            sort: Dictionary to apply sorting: {"order": "desc", "column": "lastChange"}

        Returns:
            list: List of Dictionaries
        """
        return self.client.fetch_all(
            self.post_endpoint,
            columns=columns,
            filters=filters,
            sort=sort,
            snapshot=True if self.snapshot_id else False,
            snapshot_id=self.snapshot_id,
        )

    def set_attribute_by_sn(self, serial_number, name, value):
        """
        Set a single Attribute by serial number.
        If Global will Error if already set.
        If Local will not error and either Add or Update
        :param serial_number: str: IP Fabric Unique Serial Number
        :param name: str: Attribute name (case sensitive)
        :param value: str: Attribute value (case sensitive)
        :return:
        """
        self.check_attribute_name({name})
        attribute = dict(name=name, sn=serial_number, value=value)
        if self.snapshot_id:
            return self.set_attributes_by_sn([attribute])
        resp = raise_for_status(self.client.post(self.endpoint, json=attribute))
        return resp.json()

    def set_attributes_by_sn(self, attributes: list[dict]):
        """
        Sets a list of Attributes for devices based on serial numbers.
        Will Add or Update Attributes.
        :param attributes: list: [{'sn': 'IPF SERIAL NUMBER', 'name': 'attributeName', 'value': 'SITE NAME'}]
        :return:
        """
        self.check_attribute_name({v["name"] for v in attributes})
        payload = dict(attributes=attributes)
        if self.snapshot_id:
            payload["snapshot"] = self.snapshot_id
        resp = raise_for_status(self.client.put(self.endpoint, json=payload))
        return resp.json()

    def set_site_by_sn(self, serial_number, site_name):
        """
        Set a single site by serial number
        If Global will Error if already set.
        If Local will not error and either Add or Update
        :param serial_number: str: IP Fabric Unique Serial Number
        :param site_name: str: Site name for device.
        :return:
        """
        return self.set_attribute_by_sn(serial_number, "siteName", site_name)

    def set_sites_by_sn(self, sites: list[dict]):
        """Sets a list of sites for devices based on serial numbers.

        Args:
            sites: [{'sn': 'IPF SERIAL NUMBER', 'value': 'SITE NAME'}]
        """
        [a.update(dict(name="siteName")) for a in sites]
        return self.set_attributes_by_sn(sites)

    def delete_attribute_by_sn(self, *serial_numbers):
        """Deletes attributes by Unique IP Fabric Serial Number(s)

        Args:
            serial_numbers: Serial Numbers
        """
        payload = dict(attributes=dict(sn=[str(i) for i in serial_numbers]))
        if self.snapshot_id:
            payload["snapshot"] = self.snapshot_id
        raise_for_status(self.client.request("DELETE", self.endpoint, json=payload))
        return True

    def delete_attribute_by_id(self, *attribute_ids):
        """Deletes attributes by Attribute ID(s)

        Args:
            attribute_ids: Attribute IDs
        """
        payload = dict(attributes=dict(id=[str(i) for i in attribute_ids]))
        if self.snapshot_id:
            payload["snapshot"] = self.snapshot_id
        raise_for_status(self.client.request("DELETE", self.endpoint, json=payload))
        return True

    def delete_attribute(self, *attributes):
        """
        Deletes attributes by Attribute
        :param attributes: dict: Attribute dictionaries
        :return:
        """
        return self.delete_attribute_by_id(*[str(i["id"]) for i in attributes])

    def update_local_attr_from_global(
        self,
        wait_for_load: bool = True,
        wait_for_assurance: bool = True,
        timeout: int = 60,
        retry: int = 5,
    ) -> bool:
        """
        Updates Snapshot Local Attributes based on Global.
        Returns True is successfully completed, False if the snapshot load did not complete if required.
        """
        if not self.snapshot_id:
            raise ImportError("Please initialize Attributes class with a snapshot_id.")
        recalc = self.check_sites_recalculation()
        ts = int(datetime.now().timestamp() * 1000)
        raise_for_status(self.client.post("/attributes/local/update-from-global", json={"snapshot": self.snapshot_id}))

        if (
            recalc
            and (wait_for_load or wait_for_assurance)
            and not self.client.snapshots[self.snapshot_id]._check_load_status(
                ts, wait_for_assurance, timeout, retry, "recalculate"
            )
        ):
            return False
        self.client.snapshots[self.snapshot_id]._refresh_status()
        if self.client.snapshots[self.snapshot_id].loaded and self.client.snapshot_id == self.snapshot_id:
            self.client.devices.update()
        return True

    def check_sites_recalculation(self) -> bool:
        """Checks if updating local attributes requires recalculation."""
        if not self.snapshot_id:
            raise ImportError("Please initialize Attributes class with a snapshot_id.")
        resp = raise_for_status(
            self.client.post(
                "/attributes/local/update-from-global/check-sites-recalculation", json={"snapshot": self.snapshot_id}
            )
        )
        return resp.json()["needsRecalculation"]



import logging
from typing import Any, Optional

from pydantic import BaseModel, Field, computed_field

from ipfabric.models.table import Table

logger = logging.getLogger("ipfabric")


class IpTelephony(BaseModel):
    client: Any = Field(None, exclude=True)
    sn: Optional[str] = None

    def print_tables(self):
        print(sorted([_ for _ in dir(self) if _[0] != "_" and isinstance(getattr(self, _), Table)]))

    @computed_field
    @property
    def phones(self) -> Table:
        return Table(client=self.client, endpoint="tables/inventory/phones", sn=self.sn)



import logging
from typing import Any, Optional

from pydantic import BaseModel, Field, computed_field

from ipfabric.models.table import Table

logger = logging.getLogger("ipfabric")


class Security(BaseModel):
    client: Any = Field(None, exclude=True)
    sn: Optional[str] = None

    def print_tables(self):
        print(sorted([_ for _ in dir(self) if _[0] != "_" and isinstance(getattr(self, _), Table)]))

    @computed_field
    @property
    def acl(self) -> Table:
        return Table(client=self.client, endpoint="tables/security/acl", sn=self.sn)

    @computed_field
    @property
    def acl_interface(self) -> Table:
        return Table(client=self.client, endpoint="tables/security/acl/interfaces", sn=self.sn)

    @computed_field
    @property
    def acl_global_policies(self) -> Table:
        return Table(client=self.client, endpoint="tables/security/acl/global-policies", sn=self.sn)

    @computed_field
    @property
    def dmvpn(self) -> Table:
        return Table(client=self.client, endpoint="tables/security/dmvpn", sn=self.sn)

    @computed_field
    @property
    def dhcp_snooping(self) -> Table:
        return Table(client=self.client, endpoint="tables/security/dhcp/snooping", sn=self.sn)

    @computed_field
    @property
    def dhcp_snooping_bindings(self) -> Table:
        return Table(client=self.client, endpoint="tables/security/dhcp/bindings", sn=self.sn)

    @computed_field
    @property
    def ipsec_tunnels(self) -> Table:
        return Table(client=self.client, endpoint="tables/security/ipsec/tunnels", sn=self.sn)

    @computed_field
    @property
    def ipsec_gateways(self) -> Table:
        return Table(client=self.client, endpoint="tables/security/ipsec/gateways", sn=self.sn)

    @computed_field
    @property
    def secure_ports_devices(self) -> Table:
        return Table(client=self.client, endpoint="tables/security/secure-ports/devices", sn=self.sn)

    @computed_field
    @property
    def secure_ports_interfaces(self) -> Table:
        return Table(client=self.client, endpoint="tables/security/secure-ports/interfaces", sn=self.sn)

    @computed_field
    @property
    def secure_ports_users(self) -> Table:
        return Table(client=self.client, endpoint="tables/security/secure-ports/users", sn=self.sn)

    @computed_field
    @property
    def zone_firewall_policies(self) -> Table:
        return Table(client=self.client, endpoint="tables/security/zone-firewall/policies", sn=self.sn)

    @computed_field
    @property
    def zone_firewall_interfaces(self) -> Table:
        return Table(client=self.client, endpoint="tables/security/zone-firewall/interfaces", sn=self.sn)

    @computed_field
    @property
    def lsvpn_gateways(self) -> Table:
        return Table(client=self.client, endpoint="tables/security/lsvpn/gateways", sn=self.sn)

    @computed_field
    @property
    def lsvpn_satellites(self) -> Table:
        return Table(client=self.client, endpoint="tables/security/lsvpn/satellites", sn=self.sn)



import logging
from typing import Any, Optional

from pydantic import BaseModel, Field, field_validator, model_validator

from ipfabric.tools.shared import raise_for_status, validate_ip_network_str

logger = logging.getLogger("ipfabric")


class Networks(BaseModel):
    exclude: Optional[list[str]] = Field(default_factory=list)
    include: Optional[list[str]] = Field(default_factory=list)

    @field_validator("exclude", "include")
    @classmethod
    def _verify_valid_networks(cls, v: list[str]) -> list[str]:
        return [validate_ip_network_str(_, ipv6=False) for _ in v]  # TODO: Add v6 in 7.3

    @model_validator(mode="after")
    def _verify_include_not_empty(self):
        if not self.include:
            raise ValueError("Discovery Settings Network Include list cannot be empty.")
        return self


class Discovery(BaseModel):
    client: Any = Field(exclude=True)
    _networks: Optional[Networks] = None

    def model_post_init(self, __context: Any) -> None:
        self._networks = self._get_networks()

    @property
    def networks(self):
        return self._networks

    def _get_networks(self):
        res = raise_for_status(self.client.get("settings"))
        return Networks(**res.json()["networks"])

    def update_discovery_networks(self, subnets: list, include: bool = False):
        payload = dict()
        payload["networks"] = dict()
        if include:
            payload["networks"]["include"] = subnets
            payload["networks"]["exclude"] = self.networks.exclude
        else:
            payload["networks"]["exclude"] = subnets
            payload["networks"]["include"] = self.networks.include
        res = raise_for_status(self.client.patch("settings", json=payload))
        return Networks(**res.json()["networks"])



from __future__ import annotations as _annotations

from collections import defaultdict
from os import PathLike
from pathlib import Path
from typing import Literal, List, Union

from pydantic import BaseModel, model_serializer

STATES = Literal["n/a", "n/i", "full"]


class BaseMatrix(BaseModel):
    id: str
    name: str

    def __repr__(self):
        return self.name


class Version(BaseMatrix, BaseModel):
    updated: str


class Category(BaseMatrix, BaseModel):
    versionId: str


class Task(BaseMatrix, BaseModel):
    versionId: str
    categoryId: str
    isNew: bool
    description: str


class Vendor(BaseMatrix, BaseModel):
    versionId: str
    isNew: bool


class Family(BaseMatrix, BaseModel):
    versionId: str
    vendorId: str
    isNew: bool


class Status(BaseModel):
    id: str
    familyId: str
    taskId: str
    versionId: str
    vendorId: str
    state: STATES


class Entry(BaseModel):
    category: Category
    task: Task
    vendor: Vendor
    family: Family
    status: Status

    @model_serializer
    def _dump(self):
        return dict(
            category=self.category.name,
            taskName=self.task.name,
            taskDescription=self.task.description,
            vendor=self.vendor.name,
            family=self.family.name,
            status=self.status.state,
        )


class SupportMatrix(BaseModel):
    version: Version
    matrix: list[Entry]

    @staticmethod
    def _compare(value, obj) -> bool:
        if value is None:
            return True
        elif isinstance(value, List):
            return obj.lower() in [_.lower for _ in value]
        elif isinstance(value, (str, int)):
            return obj.lower() == str(value).lower()

    def _compare_obj(self, _id, name, obj) -> bool:
        checks = {self._compare(id, obj.id)}
        checks.add(self._compare(name, obj.name))
        return False not in checks

    def filter(
        self,
        category_id: Union[str, list[str]] = None,
        category_name: Union[str, list[str]] = None,
        task_id: Union[str, list[str]] = None,
        task_name: Union[str, list[str]] = None,
        task_description: str = None,
        vendor_id: Union[str, list[str]] = None,
        vendor_name: Union[str, list[str]] = None,
        family_id: Union[str, list[str]] = None,
        family_name: Union[str, list[str]] = None,
        status: Union[STATES, list[STATES]] = None,
    ) -> SupportMatrix:
        tmp = list()
        for entry in self.matrix:
            checks = {self._compare(status, entry.status.state)}
            checks.add(self._compare_obj(category_id, category_name, entry.category))
            checks.add(self._compare_obj(task_id, task_name, entry.task))
            checks.add(self._compare_obj(vendor_id, vendor_name, entry.vendor))
            checks.add(self._compare_obj(family_id, family_name, entry.family))
            checks.add(task_description and task_description.lower() in entry.task.description.lower())
            if False not in checks:
                tmp.append(entry)
        return SupportMatrix(version=self.version, matrix=tmp)

    def to_df(self):
        try:
            import pandas as pd
        except ImportError:
            raise ImportError("Pandas not installed, please install using `pip install pandas`.")

        idx, cols = set(), set()
        data = {(entry.vendor.name, entry.family.name): defaultdict(dict) for entry in self.matrix}

        for entry in self.matrix:
            idx.add((entry.category.name, entry.task.name, entry.task.description))
            cols.add((entry.vendor.name, entry.family.name))
            data[(entry.vendor.name, entry.family.name)][
                (entry.category.name, entry.task.name, entry.task.description)
            ] = entry.status.state

        df = pd.DataFrame(
            data=data,
            columns=pd.MultiIndex.from_tuples(sorted(cols), names=["Vendor", "Family"]),
            index=pd.MultiIndex.from_tuples(sorted(idx), names=["category", "taskName", "taskDescription"]),
        )
        return df

    @staticmethod
    def _style_df(v):
        color = {
            "full": "color: green; border: 1px solid black",
            "n/i": "color: blue; border: 1px solid black",
            "n/a": "color: red; border: 1px solid black",
        }
        return [color[_] if isinstance(_, str) and _ in color else "color: #000000" for _ in v]

    def to_excel(self, filename: Union[str, PathLike]):
        filename = Path(filename).resolve().with_suffix(".xlsx").absolute()
        try:
            import openpyxl, jinja2  # noqa: E401,F401
        except ImportError:
            raise ImportError(
                "openpyxl and jinja2 not installed; please install using: "
                "`pip install openpyxl jinja2` or `pip install ipfabric[matrix]`."
            )
        self.to_df().style.apply(self._style_df, axis="columns").to_excel(filename, sheet_name=self.version.name)

    def to_csv(self, filename: Union[None, str, PathLike] = None):
        filename = Path(filename).resolve().with_suffix(".csv").absolute() if filename else None
        return self.to_df().to_csv(path_or_buf=filename)



import logging
from typing import Any, Optional, Union, Annotated

from pydantic import Field, BaseModel, StringConstraints, field_validator

logger = logging.getLogger("ipfabric")

METHOD = Annotated[str, StringConstraints(to_lower=True)]


class Scope(BaseModel):
    method: METHOD
    path: str
    api_scope_id: str = Field(alias="id")

    @field_validator("path")
    @classmethod
    def _path(cls, p):
        return p[1:]


class Token(BaseModel):
    token_id: str = Field(None, alias="id")
    description: Optional[str] = None
    role_ids: list = Field(None, alias="roleIds")


class User(BaseModel):
    username: str
    user_id: str = Field(None, alias="id")
    local: Optional[bool] = Field(None, alias="isLocal")
    sso_provider: Optional[Any] = Field(None, alias="ssoProvider")
    domains: Optional[Any] = Field(None, alias="domainSuffixes")
    role_names: Optional[list] = Field(alias="roleNames", default_factory=list)
    role_ids: list = Field(default_factory=list, alias="roleIds")
    ldap_id: Optional[Any] = Field(None, alias="ldapId")
    timezone: Optional[str] = None
    token: Optional[Token] = None
    _scopes: Optional[list[Scope]] = None
    _snapshots_settings: Optional[bool] = None

    def model_post_init(self, __context: Any) -> None:
        if self._snapshots_settings is None and self.is_admin:
            self._snapshots_settings = True

    def __repr__(self):
        return self.username

    @property
    def is_admin(self):
        return (
            True
            if (self.token and "admin" in self.token.role_ids) or (not self.token and "admin" in self.role_ids)
            else False
        )

    @property
    def snapshots_settings(self):
        """If the User/Token has access to snapshots/:key/settings Endpoint."""
        return self._snapshots_settings

    @snapshots_settings.setter
    def snapshots_settings(self, v: bool):
        self._snapshots_settings = v

    @property
    def error_msg(self):
        msg = f'User "{self.username}" '
        if self.token:
            msg += f'Token "{self.token.description}" '
        return msg

    @property
    def scopes(self) -> list[Scope]:
        return self._scopes

    @scopes.setter
    def scopes(self, scopes: list[Union[Scope, dict[str, str]]]):
        self._scopes = [Scope(**dict(_)) for _ in scopes]



import os
from typing import Optional, Union, Literal, Any, Annotated
from uuid import UUID

from pydantic import field_validator, BaseModel, Field, model_serializer, StringConstraints
from pydantic.functional_validators import AfterValidator
from pydantic_extra_types.color import ColorType, Color

from ipfabric.tools.shared import valid_snapshot
from .constants import (
    VALID_DEV_TYPES,
    DEFAULT_NETWORK,
    DEFAULT_PATHLOOKUP,
    VALID_NET_PROTOCOLS,
    VALID_PROTOCOL_LABELS,
    VALID_PATH_PROTOCOLS,
    HIDDEN_DEV_TYPES,
)

PYDANTIC_EXTRAS = os.getenv("IPFABRIC_PYDANTIC_EXTRAS", "allow")


class Style(BaseModel, extra=PYDANTIC_EXTRAS):
    color: Union[ColorType]
    pattern: Annotated[str, AfterValidator(lambda s: s.lower()), StringConstraints(pattern=r'solid|dashed|dotted')] = "solid"
    thicknessThresholds: list[int] = [2, 4, 8]

    @field_validator("color")
    @classmethod
    def _parse_color(cls, v) -> str:
        return Color(v).as_hex()


class LabelsOptions(BaseModel, extra=PYDANTIC_EXTRAS):
    wrapCenterLabel: bool = False
    wrapLinecapLabel: bool = False


class Setting(BaseModel, extra=PYDANTIC_EXTRAS):
    name: str
    style: Style
    type: Literal["protocol", "pathLookupEdge", "group"]
    visible: bool = True
    grouped: bool = True
    id: Optional[UUID] = Field(None, exclude=True)
    labelsOptions: LabelsOptions = Field(default_factory=LabelsOptions)


class EdgeSettings(Setting, BaseModel, extra=PYDANTIC_EXTRAS):
    type: Literal["protocol", "pathLookupEdge"]
    labels: list[str] = ["protocol"]


class GroupSettings(Setting, BaseModel, extra=PYDANTIC_EXTRAS):
    label: str
    children: list[EdgeSettings]
    type: Literal["group"] = "group"


class PathLookup(BaseModel, extra=PYDANTIC_EXTRAS):
    ignoredTopics: list[Literal['ACL', 'FORWARDING', 'NAT44', 'PBR', 'ZONEFW']] = Field(
        default_factory=list,
        description="List of topics to ignore.  Valid topics are in ['ACL', 'FORWARDING', 'NAT44', 'PBR', 'ZONEFW'].",
    )
    colorDetectedLoops: Optional[bool] = True


class NetworkSettings(BaseModel):
    edges: Optional[list[GroupSettings]] = None
    hiddenDeviceTypes: list[str] = Field(HIDDEN_DEV_TYPES, description="List of device types to hide.")

    def model_post_init(self, __context: Any) -> None:
        if not self.edges:
            self.edges = [GroupSettings(**edge) for edge in DEFAULT_NETWORK]

    @field_validator("hiddenDeviceTypes")
    @classmethod
    def _valid_dev_types(cls, v):
        if v and not all(d in VALID_DEV_TYPES for d in v):
            raise ValueError(f"Device Types '{v}' must be None or in {VALID_DEV_TYPES}.")
        return v

    @staticmethod
    def _update_edge(children: list[EdgeSettings], name: str, attribute: str, bool_value=False):
        for edge in children:
            if edge.name.lower() == name:
                setattr(edge, attribute, bool_value)
                return True
        return False

    def _update_group(self, name: str, attribute: str, group: bool = False, bool_value=False):
        for edge in self.edges:
            if group and isinstance(edge, GroupSettings) and edge.name.lower() == name:
                setattr(edge, attribute, bool_value)
                return True
            elif not group:
                if isinstance(edge, GroupSettings) and self._update_edge(edge.children, name, attribute, bool_value):
                    return self._update_group(edge.name.lower(), "grouped", True)
                elif isinstance(edge, EdgeSettings) and self._update_edge([edge], name, attribute, bool_value):
                    return True
        return False

    def hide_protocol(self, protocol_name: str, unhide: bool = False):
        if protocol_name.lower() in VALID_NET_PROTOCOLS:
            return self._update_group(
                VALID_NET_PROTOCOLS[protocol_name.lower()], attribute="visible", group=False, bool_value=unhide
            )
        else:
            raise KeyError(
                f"Protocol {protocol_name} does not exist.  Valid protocols are {VALID_NET_PROTOCOLS.values()}"
            )

    def hide_all_protocols(self):
        for protocol_name in VALID_NET_PROTOCOLS.values():
            self._update_group(protocol_name, attribute="visible", group=False)

    def ungroup_protocol(self, protocol_name: str):
        if protocol_name.lower() in VALID_NET_PROTOCOLS:
            return self._update_group(VALID_NET_PROTOCOLS[protocol_name.lower()], attribute="grouped", group=False)
        else:
            raise KeyError(
                f"Protocol {protocol_name} does not exist.  Valid protocols are {VALID_NET_PROTOCOLS.values()}"
            )

    def hide_group(self, group_name: str):
        group_names = [g.name.lower() for g in self.edges if isinstance(g, GroupSettings)]
        if group_name.lower() in group_names:
            return self._update_group(group_name.lower(), attribute="visible", group=True)
        else:
            raise KeyError(f"Group {group_name} does not exist.  Valid groups are {group_names}")

    def ungroup_group(self, group_name: str):
        group_names = [g.name.lower() for g in self.edges if isinstance(g, GroupSettings)]
        if group_name.lower() in group_names:
            return self._update_group(group_name.lower(), attribute="grouped", group=True)
        else:
            raise KeyError(f"Group {group_name} does not exist.  Valid groups are {group_names}")

    @staticmethod
    def _proto_label(edge: EdgeSettings, protocol_name: str, label_name: str):
        if edge.name.lower() == protocol_name:
            proto = next(x for x in VALID_PROTOCOL_LABELS[protocol_name].labels if x == label_name)
            if proto.center:
                edge.labels[0] = proto.name
            else:
                edge.labels[1] = proto.name
            return True
        return False

    def change_label(self, protocol_name: str, label_name: str):
        protocol_name, label_name = protocol_name.lower(), label_name.lower()
        if protocol_name not in VALID_NET_PROTOCOLS:
            raise KeyError(
                f"Protocol {protocol_name} does not exist.  Valid protocols are {VALID_NET_PROTOCOLS.values()}"
            )
        else:
            protocol_name = VALID_NET_PROTOCOLS[protocol_name]
        if label_name not in VALID_PROTOCOL_LABELS[protocol_name].labels:
            raise KeyError(
                f"Label {label_name} does not exist for protocol {protocol_name}.  "
                f"Valid labels for {protocol_name} are {VALID_PROTOCOL_LABELS[protocol_name].labels}"
            )
        for edge in self.edges:
            if isinstance(edge, GroupSettings):
                for child in edge.children:
                    if self._proto_label(child, protocol_name, label_name):
                        return True
            if self._proto_label(edge, protocol_name, label_name):
                return True
        return False


class PathLookupSettings(BaseModel):
    edges: Optional[list[EdgeSettings]] = None
    pathLookup: Optional[PathLookup] = Field(default_factory=PathLookup)

    def model_post_init(self, __context: Any) -> None:
        if not self.edges:
            self.edges = [EdgeSettings(**edge) for edge in DEFAULT_PATHLOOKUP]

    @property
    def protocol_priority(self):
        return {edge.name.lower(): idx for idx, edge in enumerate(self.edges)}

    def increase_priority(self, protocol_name: str):
        if protocol_name.lower() not in VALID_PATH_PROTOCOLS:
            raise KeyError(f"Protocol {protocol_name} does not exist.  Valid protocols are {VALID_PATH_PROTOCOLS}")
        current = self.protocol_priority[protocol_name]
        if current != 0:
            self.edges[current], self.edges[current - 1] = self.edges[current - 1], self.edges[current]
        return True

    def decrease_priority(self, protocol_name: str):
        if protocol_name.lower() not in VALID_PATH_PROTOCOLS:
            raise KeyError(f"Protocol {protocol_name} does not exist.  Valid protocols are {VALID_PATH_PROTOCOLS}")
        current = self.protocol_priority[protocol_name]
        if current != len(self.edges) - 1:
            self.edges[current], self.edges[current + 1] = self.edges[current + 1], self.edges[current]
        return True


class Overlay(BaseModel):
    """Set snapshotToCompare or intentRuleId, not both."""

    snapshotToCompare: Optional[Union[UUID, str]] = Field(None, description="Snapshot ID to compare.")
    intentRuleId: Optional[Union[int, str]] = Field(
        None,
        description="Intent Rule ID to overlay. Also valid: ['nonRedundantEdges', 'singlePointsOfFailure']",
    )

    @field_validator("snapshotToCompare")
    @classmethod
    def _valid_snapshot(cls, v):
        return valid_snapshot(v)

    @field_validator("intentRuleId")
    @classmethod
    def valid_intent_rule(cls, v):
        if v and v in ["nonRedundantEdges", "singlePointsOfFailure"]:
            return v
        try:
            return str(int(v))
        except ValueError:
            raise ValueError(f'"{v}" is not an Intent Rule ID or in ["nonRedundantEdges", "singlePointsOfFailure"]')

    @property
    def type(self):
        return "compare" if self.snapshotToCompare else "intent"

    @model_serializer
    def _serialize_overlay(self) -> dict:
        _ = {"intentRuleId": self.intentRuleId} if self.intentRuleId else {"snapshotToCompare": self.snapshotToCompare}
        return dict(type=self.type, **_)
